import numpy as np

from typing import Union, Dict
from copy import deepcopy
from scipy import optimize

from . import graphic
from . orbit import orbit
from . curves import lc, rv
from . surface import mesh
from . surface.temperature import interpolate_albedo
from . transform import BinarySystemProperties
from . curves import c_router
from . import (
    utils as bsutils,
    radius as bsradius,
    model
)
from . container import OrbitalPositionContainer

from .. base.error import MorphologyError
from .. base.container import SystemPropertiesContainer
from .. base.system import System
from .. base.star import Star
from .. base.curves import utils as rv_utils

from .. import settings
from .. base.types import INT, FLOAT
from .. logger import getLogger
from .. import (
    umpy as up,
    units as u,
    utils,
    const
)
from .. opt.fsolver import fsolve

logger = getLogger('binary_system.system')


class BinarySystem(System):
    """
    Class to store and calculate necessary properties of the binary system based on the user provided parameters.
    Child class of elisa.base.system.System.

    Class can be imported directly:
    ::

        >>> from elisa import BinarySystem

    After initialization, apart from the attributes already defined by the user with the arguments, user has access to
    the following attributes:

        :mass_ratio: float; secondary mass / primary mass
        :semi_major_axis: float; semi major axis of system in physical units
        :morphology: str; morphology of the system:

                      :`detached`: both components are not filling their respective Roche lobes,
                      :`semi-detached`: one of the components is filling its Roche lobe,
                      :`double-contact`: both components fill their Roche lobes,
                      :`over-contact`: components are physically connected with a ''neck'')

    `BinarySystem' requires instances of elisa.base.star.Star in `primary` and `secondary` argument with following
    mandatory arguments:

        :param mass: float; If mass is int, np.int, float, np.float, program assumes solar mass as it's unit.
                            If mass astropy.unit.quantity.Quantity instance, program converts it to default units.
        :param t_eff: float; Accepts value in any temperature unit. If your input is without unit,
                             function assumes that supplied value is in K.
        :param surface_potential: float; generalized surface potential (Wilson 79)
        :param synchronicity: float; synchronicity F (omega_rot / omega_orb), equals 1 for synchronous rotation

    following optional arguments are also available:

        :param metallicity: float; log[M/H] default value is 0.0
        :param gravity_darkening: float; gravity darkening factor, if not supplied, it is interpolated
                                         from Claret 2003 based on t_eff

        :param albedo: float; surface albedo, value from <0, 1> interval, if not supplied,
                              Claret 2001 will be used for interpolation
        :param limb_darkening_coefficients: Union[float, dict]; optional limb darkening coefficients
                                            used for the whole star useful in case the modelled star is outside the
                                            supported range of atmospheric parameters. Limb darkening coefficients can
                                            be supplied as dict {passband: ld_coefs}. If unused, elisa will
                                            interpolate the values from supplied limb-darkening tables.

    Each component instance will after initialization contain following attributes:

        :critical_surface_potential: float; potential of the star required to fill its Roche lobe
        :equivalent_radius: float; radius of a sphere with the same volume as a component (in SMA units)
        :filling_factor: float: calculated as (Omega_{inner} - Omega) / (Omega_{inner} - Omega_{outter})

                            :filling factor < 0: component does not fill its Roche lobe
                            :filling factor = 0: component fills preciselly its Roche lobe
                            :1 > filling factor > 0: component overflows its Roche lobe
                            :filling factor = 1: upper boundary of the filling factor, higher value would lead to
                                                 the mass loss trough Lagrange point L2

        Radii at periastron (in SMA units)
            :polar_radius: float; radius of a star towards the pole of the star
            :side_radius: float; radius of a star in the direction perpendicular to the pole
                                 and direction of a companion
            :backward_radius: float; radius of a star in the opposite direction as the binary companion
            :forward_radius: float; radius of a star towards the binary companion,
                                    returns numpy.nan if the system is over-contact

    The BinarySystem can be initialized either by using valid class arguments, e.g.:
    ::

        >>> from elisa import BinarySystem
        >>> from elisa import Star
        >>> from astropy import units as u

        >>> primary = Star(
        >>>     mass=2.15 * u.solMass,
        >>>     surface_potential=3.6,
        >>>     synchronicity=1.0,
        >>>      t_eff=10000 * u.K,
        >>>     gravity_darkening=1.0,
        >>>     discretization_factor=5,  # angular size (in degrees) of the surface elements
        >>>     albedo=0.6,
        >>>     metallicity=0.0,
        >>> )

        >>> secondary = Star(
        >>>     mass=0.45 * u.solMass,
        >>>     surface_potential=5.39,
        >>>     synchronicity=1.0,
        >>>     t_eff=8000 * u.K,
        >>>     gravity_darkening=1.0,
        >>>     albedo=0.6,
        >>>     metallicity=0,
        >>> )
        >>>
        >>> bs = BinarySystem(
        >>>     primary=primary,
        >>>     secondary=secondary,
        >>>     argument_of_periastron=58 * u.deg,
        >>>     gamma=-30.7 * u.km / u.s,
        >>>     period=2.5 * u.d,
        >>>     eccentricity=0.0,
        >>>     inclination=85 * u.deg,
        >>>     primary_minimum_time=2440000.00000 * u.d,
        >>>     phase_shift=0.0,
        >>>     distance=162 * u.pc
        >>> )

    or by using the BinarySystem.from_json(<dict>) function that accepts various parameter combination in form of
    dictionary such as:
    ::

        >>> data = {
        >>>     "system": {
        >>>         "inclination": 90.0,
        >>>         "period": 10.1,
        >>>         "argument_of_periastron": 90.0,
        >>>         "gamma": "0.0 m / s",  # you can define quantity using string representation of the astropy units
        >>>         "eccentricity": 0.3,
        >>>         "primary_minimum_time": 0.0,
        >>>         "phase_shift": 0.0,
        >>>         "distance": "152 pc"
        >>>     },
        >>>     "primary": {
        >>>         "mass": 2.15,
        >>>         "surface_potential": 3.6,
        >>>         "synchronicity": 1.0,
        >>>         "t_eff": 10000.0,
        >>>         "gravity_darkening": 1.0,
        >>>         "discretization_factor": 5,
        >>>         "albedo": 1.0,
        >>>         "metallicity": 0.0,
        >>>         "atmosphere": "ck04"
        >>>     },
        >>>     "secondary": {
        >>>         "mass": 0.45,
        >>>         "surface_potential": 5.39,
        >>>         "synchronicity": 1.0,
        >>>         "t_eff": 8000.0,
        >>>         "gravity_darkening": 1.0,
        >>>         "albedo": 1.0,
        >>>         "metallicity": 0.0,
        >>>         "atmosphere": "black_body"
        >>>     }
        >>> }
        >>>
        >>> binary = BinarySystem.from_json(data)

    See documentation for `from_json` method for details.

    The orbit of the binary system can be modelled using function
    `calculate_orbital_motion(phases)`. E.g.:
    ::

        >>> binary.calculate_orbital_motion(np.linspace(0, 1))

    The class contains substantial plotting capability in the BinarySystem.plot module that contains following
    functions (further info in: see elisa.binary_system.graphics.plot):

        - orbit(args): plots an orbit of the binary system
        - equipotential(args): xy, yz, zx cross-sections of equipotential surface
        - mesh(args): 3D mesh (scatter) plot of the surface points
        - wireframe(args): wire frame model of the selected system components
        - surface(args): plot models of the binary components with various surface
                         colormaps (gravity_acceleration, temperature, radiance, ...)

    Plot function can be called as function of the plot module. E.g.:
    ::

        >>> binary.plot.surface(phase=0.1, colormap='temperature'))

    Similarly, an animation of the orbital motion can be produced using BinarySystem.animation module and its function
    `orbital_motion(*args)`.


    List of valid input system arguments:

    :param primary: elisa.base.star.Star; instance of primary component
    :param secondary: elisa.base.star.Star; instance of secondary component
    :param inclination: Union[float, astropy.unit.quantity.Quantity]; Inclination of the system.
                        If unit is not supplied, value in degrees is assumed.
    :param period: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]; Orbital period of binary
                   star system. If unit is not specified, default period unit is assumed (days).
    :param eccentricity: Union[(numpy.)int, (numpy.)float]; from <0, 1> interval
    :param argument_of_periastron: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity];
    :param gamma: Union[float, astropy.unit.quantity.Quantity]; Center of mass velocity.
                  Expected type is astropy.units.quantity.Quantity, numpy.float or numpy.int
                  otherwise TypeError will be raised. If unit is not specified, default velocity unit is assumed (m/s).
    :param phase_shift: float; Phase shift of the primary eclipse with respect to the ephemeris.
                               true_phase is used during calculations, where: true_phase = phase + phase_shift.;
    :param primary_minimum_time: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity];
    :param additional_light: float; fraction of light that does not originate from the `BinarySystem`
    :param distance: float: distance between system and the observer
    """

    MANDATORY_KWARGS = ['inclination', 'period', 'eccentricity', 'argument_of_periastron']
    OPTIONAL_KWARGS = ['gamma', 'phase_shift', 'additional_light', 'primary_minimum_time', 'distance']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    STAR_MANDATORY_KWARGS = ['mass', 't_eff', 'surface_potential', 'synchronicity']
    STAR_OPTIONAL_KWARGS = ['metallicity', 'gravity_darkening', 'albedo']
    STAR_ALL_KWARGS = STAR_MANDATORY_KWARGS + STAR_OPTIONAL_KWARGS

    def __init__(self, primary, secondary, name=None, **kwargs):
        # initial validity checks
        utils.invalid_kwarg_checker(kwargs, BinarySystem.ALL_KWARGS, self.__class__)
        utils.check_missing_kwargs(BinarySystem.MANDATORY_KWARGS, kwargs, instance_of=BinarySystem)
        self.object_params_validity_check(dict(primary=primary, secondary=secondary), self.STAR_MANDATORY_KWARGS)
        kwargs: Dict = self.transform_input(**kwargs)

        super(BinarySystem, self).__init__(name, **kwargs)

        logger.info(f"initialising object {self.__class__.__name__}")
        logger.debug(f"setting properties of components of class instance {self.__class__.__name__}")

        # graphic related properties
        self.plot = graphic.plot.Plot(self)
        self.animation = graphic.animation.Animation(self)

        # components
        self.primary: Star = primary
        self.secondary: Star = secondary
        self._components: Dict[str, Star] = dict(primary=self.primary, secondary=self.secondary)

        # default values of properties
        self.orbit: Union[None, orbit.Orbit] = None
        self.period: float = np.nan
        self.eccentricity: float = np.nan
        self.argument_of_periastron: float = np.nan
        self.primary_minimum_time: float = 0.0
        self.phase_shift: float = 0.0
        self.gamma: float = 0.0
        self.mass_ratio: float = self.secondary.mass / self.primary.mass

        # set attributes and test whether all parameters were initialized
        # we already ensured that all kwargs are valid and all mandatory kwargs are present so lets set class attributes
        self.init_properties(**kwargs)

        # calculation of dependent parameters
        logger.debug("computing semi-major axis")
        self.semi_major_axis: float = self.calculate_semi_major_axis()

        # orbit initialisation (initialise class Orbit from given BinarySystem parameters)
        self.init_orbit()

        # setup critical surface potentials in periastron
        logger.debug("setting up critical surface potentials of components in periastron")
        self.setup_periastron_critical_potential()

        logger.debug("setting up morphological classification of binary system")
        self.morphology: str = self.compute_morphology()

        self.setup_components_radii(
            components_distance=self.orbit.periastron_distance,
            calculate_equivalent_radius=True
        )
        self.setup_betas()
        self.setup_albedos()
        self.assign_pulsations_amplitudes(normalisation_constant=self.semi_major_axis)

        # adjust and setup discretization factor if necessary
        self.setup_discretisation_factor()

        # setting common reference to emphemeris
        self.t0: float = self.primary_minimum_time

    @property
    def default_input_units(self):
        """
        Returns set of default units of intialization parameters, in case, when provided without an units.

        :return: elisa.units.DefaultBinarySystemInputUnits;
        """
        return u.DefaultBinarySystemInputUnits

    @property
    def default_internal_units(self):
        """
        Returns set of internal units of system parameters.

        :return: elisa.units.DefaultBinarySystemUnits;
        """
        return u.DefaultBinarySystemUnits

    @classmethod
    def from_json(cls, data, _verify=True, _kind_of=None):
        """
        Create instance of BinarySystem from JSON in form such as::

            {
              "system": {
                "inclination": 90.0,
                "period": 10.1,
                "argument_of_periastron": "90.0 deg",  # string representation of astropy quntity is also valid
                "gamma": 0.0,
                "eccentricity": 0.3,
                "primary_minimum_time": 0.0,
                "phase_shift": 0.0,
                "distance": 155
              },
              "primary": {
                "mass": 2.0,
                "surface_potential": 7.1,
                "synchronicity": 1.0,
                "t_eff": 6500.0,
                "gravity_darkening": 1.0,
                "discretization_factor": 5,
                "albedo": 1.0,
                "metallicity": 0.0,
                "atmosphere": "ck04"
              },
              "secondary": {
                "mass": 2.0,
                "surface_potential": 7.1,
                "synchronicity": 1.0,
                "t_eff": 6500.0,
                "gravity_darkening": 1.0,
                "discretization_factor": 5,
                "albedo": 1.0,
                "metallicity": 0.0,
                "atmosphere": "black_body"
              }
            }

            or

            {
              "system": {
                "inclination": 90.0,
                "period": 10.1,
                "argument_of_periastron": 90.0,
                "gamma": 0.0,
                "eccentricity": 0.3,
                "primary_minimum_time": 0.0,
                "phase_shift": 0.0,
                "semi_major_axis": 10.5,
                "mass_ratio": 0.5,
                "distance": "125 pc"
              },
              "primary": {
                "surface_potential": 7.1,
                "synchronicity": 1.0,
                "t_eff": 6500.0,
                "gravity_darkening": 1.0,
                "discretization_factor": 5,
                "albedo": 1.0,
                "metallicity": 0.0,
                "atmosphere": "black_body"
              },
              "secondary": {
                "surface_potential": 7.1,
                "synchronicity": 1.0,
                "t_eff": 6500.0,
                "gravity_darkening": 1.0,
                "discretization_factor": 5,
                "albedo": 1.0,
                "metallicity": 0.0,
                "atmosphere": "black_body"
              }
            }

        Default units (when unit is not specified as string)::

             {
                "inclination": [degrees],
                "period": [days],
                "argument_of_periastron": [degrees],
                "gamma": [m/s],
                "eccentricity": [dimensionless],
                "primary_minimum_time": [d],
                "phase_shift": [dimensionless],
                "distance": [pc]
                "mass": [solMass],
                "surface_potential": [dimensionless],
                "synchronicity": [dimensionless],
                "t_eff": [K],
                "gravity_darkening": [dimensionless],
                "discretization_factor": [degrees],
                "albedo": [dimensionless],
                "metallicity": [dimensionless],
                "semi_major_axis": [solRad],
                "mass_ratio": [dimensionless],
                "limb_darkening_coefficients": [dimensionless]
            }

        :return: elisa.binary_system.system.BinarySystem;
        """
        data_cp = deepcopy(data)
        if _verify:
            bsutils.validate_binary_json(data_cp)

        kind_of = _kind_of or bsutils.resolve_json_kind(data_cp)
        if kind_of in ["community"]:
            data_cp = bsutils.transform_json_community_to_std(data_cp)

        primary, secondary = Star(**data_cp["primary"]), Star(**data_cp["secondary"])
        return cls(primary=primary, secondary=secondary, **data_cp["system"])

    @classmethod
    def from_fit_results(cls, results, atmosphere=None, limb_darkening_coefficients=None):
        """
        Building binary system from standard fit results format.

        :param results: Dict; {'component': {'param_name': {'value': value, fixed: ...}}}
        :param atmosphere: dict; atmosphere model for each component eg. 'ck04' or 'bb'
        :param limb_darkening_coefficients: dict; custom limb-darkening coefficents for each component and passband
        :return: elisa.binary_system.system.BinarySystem;
        """
        extra_parameters = {'atmosphere': atmosphere, 'limb_darkening_coefficients': limb_darkening_coefficients}

        data = dict()
        for key, component in results.items():
            if key == 'r_squared':
                continue
            data[key] = dict()

            for param, content in component.items():
                if param in ['spots', 'pulsations']:
                    features = list()
                    for idx, feature in enumerate(content):
                        features.append(dict())
                        for f_param, f_content in feature.items():
                            if f_param == 'label':
                                continue
                            features[idx][f_param] = f_content['value']
                    data[key][param] = features
                else:
                    data[key][param] = content['value']

            for extra_param, val in extra_parameters.items():
                if val is not None and val.get(key) is not None:
                    data[key][extra_param] = val[key]

        return BinarySystem.from_json(data=data)

    def build_container(self, phase=None, time=None, build_pulsations=True):
        """
        Function returns `OrbitalPositionContainer` with fully built model binary system at
        user-defined photometric phase or time of observation.

        :param time: float; JD, Julian-Date
        :param phase: float; photometric phase
        :param build_pulsations: bool; whether build pulsation or not
        :return: elisa.binary_system.container.OrbitalPositionContainer;
        """
        if phase is not None and time is not None:
            raise ValueError('Please specify whether you want to build your container '
                             'EITHER at given photometric `phase` or at given `time`.')
        phase = phase if time is None else utils.jd_to_phase(time, period=self.period, t0=self.t0)

        position = self.calculate_orbital_motion(input_argument=phase, return_nparray=False, calculate_from='phase')[0]
        orbital_position_container = OrbitalPositionContainer.from_binary_system(self, position)
        orbital_position_container.build(build_pulsations=build_pulsations)

        logger.info(f'Orbital position container was successfully built at photometric phase {phase:.2f}.')
        return orbital_position_container

    def init(self):
        """
        Function to reinitialize BinarySystem class instance after changing parameter(s) of binary system.
        """
        # reinitialize components (in case when value in component instance was changed)
        for component in settings.BINARY_COUNTERPARTS:
            getattr(self, component).init()

        self.__init__(primary=self.primary, secondary=self.secondary, **self.kwargs_serializer())

    @property
    def components(self):
        """
        Return components object in Dict[str, elisa.base.star.Star].

        :return: Dict[str, elisa.base.star.Star];
        """
        return self._components

    def properties_serializer(self):
        """
        Return binary system properties in form of a JSON file.

        :return: Dict; {`primary`: {}, `secondary`: {}, `system`: {}}
        """
        props = BinarySystemProperties.transform_input(**self.kwargs_serializer())
        props.update({
            "semi_major_axis": self.semi_major_axis,
            "morphology": self.morphology,
            "mass_ratio": self.mass_ratio
        })
        return props

    def to_properties_container(self):
        return SystemPropertiesContainer(**self.properties_serializer())

    def init_orbit(self):
        """
        Orbit class in binary system.
        """
        logger.debug(f"re/initializing orbit in class instance {self.__class__.__name__} / {self.name}")
        orbit_kwargs = {key: getattr(self, key) for key in orbit.Orbit.ALL_KWARGS}
        self.orbit = orbit.Orbit(**orbit_kwargs)

    def is_eccentric(self):
        """
        Resolve whether system is eccentric.

        :return: bool;
        """
        return self.eccentricity > 0

    def is_synchronous(self):
        """
        Resolve whether system is synchronous (consider synchronous system
        if sychnronicity of both components is equal to 1).

        :return: bool;
        """
        return (self.primary.synchronicity == 1) & (self.secondary.synchronicity == 1)

    def calculate_semi_major_axis(self):
        """
        Calculates length of semi major axis using 3rd kepler default_law.

        :return: float;
        """
        period = np.float64((self.period * u.DefaultBinarySystemUnits.system.period).to(u.TIME_UNIT))
        return (const.G * (self.primary.mass + self.secondary.mass) * period ** 2 / (4 * const.PI ** 2)) ** (1.0 / 3)

    def compute_morphology(self):
        """
        Setup binary star class property `morphology`.
        Determines the morphology based on current system parameters
        and setup `morphology` parameter of `self` system instance.

        :return: str; morphology of the system (`detached`, `semi-detached`, `double-contact`, `over-contact`)
        """
        __PRECISSION__ = 1e-8
        __MORPHOLOGY__ = None
        if (self.primary.synchronicity == 1 and self.secondary.synchronicity == 1) and self.eccentricity == 0.0:
            lp = self.libration_potentials()

            self.primary.filling_factor = self.compute_filling_factor(self.primary.surface_potential, lp)
            self.secondary.filling_factor = self.compute_filling_factor(self.secondary.surface_potential, lp)

            if ((1 > self.secondary.filling_factor > 0) or (1 > self.primary.filling_factor > 0)) and \
                    (abs(self.primary.filling_factor - self.secondary.filling_factor) > __PRECISSION__):
                msg = "Detected over-contact binary system, but potentials of components are not the same."
                raise MorphologyError(msg)
            if self.primary.filling_factor > 1 or self.secondary.filling_factor > 1:
                raise MorphologyError("Non-Physical system: primary_filling_factor or "
                                      "secondary_filling_factor is greater then 1. "
                                      "Filling factor is obtained as following:"
                                      "(Omega_{inner} - Omega) / (Omega_{inner} - Omega_{outter})")

            if (abs(self.primary.filling_factor) < __PRECISSION__ and self.secondary.filling_factor < 0) or \
                    (self.primary.filling_factor < 0 and abs(self.secondary.filling_factor) < __PRECISSION__) or \
                    (abs(self.primary.filling_factor) < __PRECISSION__ and abs(self.secondary.filling_factor)
                     < __PRECISSION__):
                __MORPHOLOGY__ = "semi-detached"
            elif self.primary.filling_factor < 0 and self.secondary.filling_factor < 0:
                __MORPHOLOGY__ = "detached"
            elif 1 >= self.primary.filling_factor > 0:
                __MORPHOLOGY__ = "over-contact"
            elif self.primary.filling_factor > 1 or self.secondary.filling_factor > 1:
                raise MorphologyError("Non-Physical system: potential of components is to low.")

        else:
            self.primary.filling_factor, self.secondary.filling_factor = None, None
            if (abs(self.primary.surface_potential - self.primary.critical_surface_potential) < __PRECISSION__) and \
                    (abs(
                        self.secondary.surface_potential - self.secondary.critical_surface_potential) < __PRECISSION__):
                __MORPHOLOGY__ = "double-contact"

            elif (not (not (abs(
                    self.primary.surface_potential - self.primary.critical_surface_potential) < __PRECISSION__) or not (
                    self.secondary.surface_potential > self.secondary.critical_surface_potential))) or \
                    ((abs(
                        self.secondary.surface_potential - self.secondary.critical_surface_potential) < __PRECISSION__)
                     and (self.primary.surface_potential > self.primary.critical_surface_potential)):
                __MORPHOLOGY__ = "semi-detached"

            elif (self.primary.surface_potential > self.primary.critical_surface_potential) and (
                    self.secondary.surface_potential > self.secondary.critical_surface_potential):
                __MORPHOLOGY__ = "detached"

            else:
                raise MorphologyError("Non-Physical system. Change stellar parameters.")
        return __MORPHOLOGY__

    def setup_discretisation_factor(self):
        """
        Adjusting discretization factors of both components to have roughly similar sizes.
        If none of the components have their discretization factors set, smaller
        component is adjusted according to bigger. If secondary discretization factor
        was not set, it will be now with respect to primary component.
        """
        def _adjust_alpha(adj_component, ref_comp):
            return ref_comp.discretization_factor * \
                   (ref_comp.equivalent_radius / adj_component.equivalent_radius) * \
                   (ref_comp.t_eff / adj_component.t_eff) ** 2

        adj_comp, adj, ref = None, None, None
        # if both components are not specified, alpha of smaller component is adjusted to the bigger component
        if self.primary.kwargs.get('discretization_factor') is None \
                and self.secondary.kwargs.get('discretization_factor') is None:
            if self.secondary.equivalent_radius * self.secondary.t_eff ** 2 < \
                    self.primary.equivalent_radius * self.primary.t_eff ** 2:
                adj, ref = self.secondary, self.primary
                adj_comp = 'secondary'
            else:
                adj, ref = self.primary, self.secondary
                adj_comp = 'primary'
        # if only one alpha is supplied, the second is adjusted
        elif self.secondary.kwargs.get('discretization_factor') is None:
            adj, ref = self.secondary, self.primary
            adj_comp = 'secondary'
        elif self.primary.kwargs.get('discretization_factor') is None:
            adj, ref = self.primary, self.secondary
            adj_comp = 'primary'

        # adjust discretization factor for given component
        if adj_comp is not None:
            adj.discretization_factor = _adjust_alpha(adj, ref)

            if adj.discretization_factor > np.radians(settings.MAX_DISCRETIZATION_FACTOR):
                adj.discretization_factor = np.radians(settings.MAX_DISCRETIZATION_FACTOR)
            if adj.discretization_factor < np.radians(settings.MIN_DISCRETIZATION_FACTOR):
                adj.discretization_factor = np.radians(settings.MIN_DISCRETIZATION_FACTOR)

            logger.info(f"setting discretization factor of {adj_comp} component to "
                        f"{up.degrees(adj.discretization_factor):.2f} "
                        f"according to discretization factor of the companion.")

        # adjust discretization factor for spots
        for component in settings.BINARY_COUNTERPARTS:
            instance: Star = getattr(self, component)
            if instance.has_spots():
                for spot in instance.spots.values():
                    if not spot.kwargs.get('discretization_factor'):
                        spot.discretization_factor = instance.discretization_factor

    def transform_input(self, **kwargs):
        """
        Transform and validate input kwargs.

        :param kwargs: Dict;
        :return: Dict;
        """
        return BinarySystemProperties.transform_input(**kwargs)

    def setup_periastron_critical_potential(self):
        """
        Compute and set critical surface potential for both components.
        Critical surface potential is for component defined as potential when component fill its Roche lobe.
        """
        for component, instance in self.components.items():
            cp = self.critical_potential(component=component, components_distance=1.0 - self.eccentricity)
            setattr(instance, "critical_surface_potential", cp)

    def critical_potential(self, component, components_distance):
        """
        Return a critical potential for target component.

        :param component: str; define target component to compute critical potential; `primary` or `secondary`
        :param components_distance: numpy.float;
        :return: numpy.float;
        """
        synchronicity = self.primary.synchronicity if component == 'primary' else self.secondary.synchronicity
        return self.critical_potential_static(component, components_distance, self.mass_ratio, synchronicity)

    @staticmethod
    def critical_potential_static(component, components_distance, mass_ratio, synchronicity):
        """
        Static method for calculation of critical potential for EB system component.

        :param component: str; define target component to compute critical potential; `primary` or `secondary`
        :param components_distance: components_distance: numpy.float;
        :param mass_ratio: float;
        :param synchronicity: float;
        :return: numpy.float;
        """
        args1 = synchronicity, mass_ratio, components_distance
        args2 = args1 + (0.0, const.HALF_PI)
        solver_err = ValueError("Iteration process to solve critical potential seems "
                                "to lead nowhere (critical potential solver has failed).")

        if component == 'primary':
            solution = optimize.newton(model.primary_potential_derivative_x, x0=1e-6, args=args1, tol=1e-12)
            if not up.isnan(solution):
                precalc_args = model.pre_calculate_for_potential_value_primary(*args2)
                args = (mass_ratio,) + precalc_args
                return abs(model.potential_value_primary(solution, *args))
            else:
                raise solver_err

        elif component == 'secondary':
            solution = optimize.newton(model.secondary_potential_derivative_x, x0=1e-6, args=args1, tol=1e-12)
            if not up.isnan(solution):
                precalc_args = model.pre_calculate_for_potential_value_secondary(*args2)
                args = (mass_ratio,) + precalc_args
                return abs(model.potential_value_secondary(components_distance - solution, *args))
            else:
                raise solver_err

        else:
            raise ValueError("Parameter `component` has incorrect value. Use `primary` or `secondary`.")

    def libration_potentials(self):
        """
        Return potentials in L3, L1, L2 respectively.

        :return: List; [Omega(L3), Omega(L1), Omega(L2)];
        """
        return self.libration_potentials_static(self.orbit.periastron_distance, self.mass_ratio)

    @staticmethod
    def libration_potentials_static(periastron_distance, mass_ratio):
        """
        Static version of `libration_potentials` where potentials in L3, L1, L2 are calculated.

        :param periastron_distance: float;
        :param mass_ratio: float;
        :return: List;
        """
        def _potential(radius):
            theta, d = const.HALF_PI, periastron_distance
            if isinstance(radius, (float, int, FLOAT, INT)):
                radius = [radius]
            elif not isinstance(radius, tuple([list, np.array])):
                raise ValueError("Incorrect value of variable `radius`.")

            p_values = []
            for r in radius:
                phi, r = (0.0, r) if r >= 0 else (const.PI, abs(r))

                block_a = 1.0 / r
                block_b = mass_ratio / (up.sqrt(up.power(d, 2) + up.power(r, 2) - (
                        2.0 * r * up.cos(phi) * up.sin(theta) * d)))
                block_c = (mass_ratio * r * up.cos(phi) * up.sin(theta)) / (up.power(d, 2))
                block_d = 0.5 * (1 + mass_ratio) * up.power(r, 2) * (
                        1 - up.power(up.cos(theta), 2))

                p_values.append(block_a + block_b - block_c + block_d)
            return p_values

        lagrangian_points = BinarySystem.lagrangian_points_static(periastron_distance, mass_ratio)
        return _potential(lagrangian_points)

    def lagrangian_points(self):
        """
        Compute Lagrangian points for current system parameters.

        :return: List; x-valeus of libration points [L3, L1, L2] respectively
        """
        return self.lagrangian_points_static(self.orbit.periastron_distance, self.mass_ratio)

    @staticmethod
    def lagrangian_points_static(periastron_distance, mass_ratio):
        """
        Static version of `lagrangian_points` that returns Lagrangian points for current system parameters.

        :param periastron_distance: float;
        :param mass_ratio: float;
        :return: List; x-valeus of libration points [L3, L1, L2] respectively
        """

        def _potential_dx(x, *args):
            """
            General potential derivatives in x when::

                primary.synchornicity = secondary.synchronicity = 1.0
                eccentricity = 0.0
            :param x: (numpy.)float
            :param args: Tuple; periastron distance of components
            :return: (numpy.)float
            """
            d, = args
            r_sqr, rw_sqr = x ** 2, (d - x) ** 2
            return - (x / r_sqr ** (3.0 / 2.0)) + ((mass_ratio * (d - x)) / rw_sqr ** (
                    3.0 / 2.0)) + (mass_ratio + 1) * x - mass_ratio / d ** 2

        xs = np.linspace(- periastron_distance * 3.0, periastron_distance * 3.0, 100)

        args_val = periastron_distance,
        round_to = 10
        points, lagrange = [], []

        for x_val in xs:
            try:
                # if there is no valid value (in case close to x=0.0, _potential_dx diverge)
                old_settings = np.seterr(divide='raise', invalid='raise')
                _potential_dx(round(x_val, round_to), *args_val)
                np.seterr(**old_settings)
            except Exception as e:
                logger.debug(f"invalid value passed to potential, exception: {str(e)}")
                continue

            try:
                solution, _, ier, _ = fsolve(_potential_dx, x_val, full_output=True, args=args_val, xtol=1e-12)
                if ier == 1:
                    if round(solution[0], 5) not in points:
                        try:
                            value_dx = abs(round(_potential_dx(solution[0], *args_val), 4))
                            use = True if value_dx == 0 else False
                        except Exception as e:
                            logger.debug(f"skipping solution for x: {x_val} due to exception: {str(e)}")
                            use = False

                        if use:
                            points.append(round(solution[0], 5))
                            lagrange.append(solution[0])
                            if len(lagrange) == 3:
                                break
            except Exception as e:
                logger.debug(f"solution for x: {x_val} lead to nowhere, exception: {str(e)}")
                continue

        return sorted(lagrange) if mass_ratio < 1.0 else sorted(lagrange, reverse=True)

    def compute_equipotential_boundary(self, components_distance, plane):
        """
        Compute a equipotential boundary of components (crossection of Hill plane).

        :param components_distance: (numpy.)float;
        :param plane: str; to compute crossection with plane `xy`, `yz` or `zx`
        :return: Tuple; (numpy.array, numpy.array) - (xs, ys) in Cartesian 2-D plane
        """

        components = ['primary', 'secondary']
        points_primary, points_secondary = [], []
        fn_map = {'primary': (model.potential_primary_fn, model.pre_calculate_for_potential_value_primary),
                  'secondary': (model.potential_secondary_fn, model.pre_calculate_for_potential_value_secondary)}

        angles = np.linspace(0, const.FULL_ARC, 500, endpoint=True)
        for component in components:
            component_instance = getattr(self, component)
            synchronicity = component_instance.synchronicity

            for angle in angles:
                if utils.is_plane(plane, 'xy'):
                    args, use = (synchronicity, self.mass_ratio, components_distance, angle, const.HALF_PI), False
                elif utils.is_plane(plane, 'yz'):
                    args, use = (synchronicity, self.mass_ratio, components_distance, const.HALF_PI, angle), False
                elif utils.is_plane(plane, 'zx'):
                    args, use = (synchronicity, self.mass_ratio, components_distance, 0.0, angle), False
                else:
                    raise ValueError('Invalid choice of crossection plane, use only: `xy`, `yz`, `zx`.')

                scipy_solver_init_value = np.array([components_distance / 10000.0])
                aux_args = (self.mass_ratio,) + fn_map[component][1](*args)
                args = (aux_args, component_instance.surface_potential)
                solution, _, ier, _ = fsolve(fn_map[component][0], scipy_solver_init_value,
                                             full_output=True, args=args, xtol=1e-12)

                # check for regular solution
                if ier == 1 and not up.isnan(solution[0]):
                    solution = solution[0]
                    if 30 >= solution >= 0:
                        use = True
                else:
                    continue

                if use:
                    if utils.is_plane(plane, 'yz'):
                        if component == 'primary':
                            points_primary.append([solution * up.sin(angle), solution * up.cos(angle)])
                        elif component == 'secondary':
                            points_secondary.append([solution * up.sin(angle), solution * up.cos(angle)])
                    elif utils.is_plane(plane, 'xz'):
                        if component == 'primary':
                            points_primary.append([solution * up.sin(angle), solution * up.cos(angle)])
                        elif component == 'secondary':
                            points_secondary.append([- (solution * up.sin(angle) - components_distance),
                                                     solution * up.cos(angle)])
                    else:
                        if component == 'primary':
                            points_primary.append([solution * up.cos(angle), solution * up.sin(angle)])
                        elif component == 'secondary':
                            points_secondary.append([- (solution * up.cos(angle) - components_distance),
                                                     solution * up.sin(angle)])

        return np.array(points_primary), np.array(points_secondary)

    def get_positions_method(self):
        """
        Return method to use for orbital motion computation.

        :return: callable;
        """
        return self.calculate_orbital_motion

    def calculate_orbital_motion(self, input_argument=None, return_nparray=False, calculate_from='phase'):
        """
        Calculate orbital motion for current system parameters and supplied phases or azimuths.

        :param calculate_from: str; 'phase' or 'azimuths' parameter based on which orbital motion should be calculated
        :param return_nparray: bool; if True positions in form of numpy arrays will be also returned
        :param input_argument: numpy.array;
        :return: Tuple[List[NamedTuple: elisa.const.Position], List[Integer]] or
                 List[NamedTuple: elisa.const.Position]
        """
        input_argument = np.array([input_argument]) if np.isscalar(input_argument) else input_argument
        orbital_motion = self.orbit.orbital_motion(phase=input_argument) if calculate_from == 'phase' \
            else self.orbit.orbital_motion_from_azimuths(azimuth=input_argument)
        idx = up.arange(np.shape(input_argument)[0], dtype=INT)
        positions = np.hstack((idx[:, np.newaxis], orbital_motion))
        # return retval, positions if return_nparray else retval
        return positions if return_nparray else [const.Position(*p) for p in positions]

    def calculate_components_radii(self, components_distance):
        """
        Calculate component radii.
        Use methods to calculate polar, side, backward and if not W UMa also
        forward radius and assign to component instance.

        :param components_distance: float; distance of components in SMA unit
        :return: Dict[str, Dict[str, float]];
        """
        fns = [bsradius.calculate_polar_radius, bsradius.calculate_side_radius, bsradius.calculate_backward_radius]
        components = settings.BINARY_COUNTERPARTS

        if self.eccentricity == 0.0:
            corrected_potential = {component: getattr(self, component).surface_potential for component in components}
        else:
            corrected_potential = self.correct_potentials(distances=np.array([components_distance, ]))
            corrected_potential = {component: corrected_potential[component][0] for component in components}

        radii = dict(primary=dict(), secondary=dict())
        for component in components:
            instance: Star = getattr(self, component)

            kwargs = dict(synchronicity=instance.synchronicity,
                          mass_ratio=self.mass_ratio,
                          components_distance=components_distance,
                          surface_potential=corrected_potential[component],
                          component=component)

            for fn in fns:
                logger.debug(f'initialising {" ".join(str(fn.__name__).split("_")[1:])} '
                             f'for {component} component')

                param_name = f'{"_".join(str(fn.__name__).split("_")[1:])}'
                r = fn(**kwargs)
                radii[component][param_name] = r

            if self.morphology != 'over-contact':
                radii[component]['forward_radius'] = bsradius.calculate_forward_radius(**kwargs)

        return radii

    def setup_components_radii(self, components_distance, calculate_equivalent_radius=True):
        """
        Setup component radii.
        Use methods to calculate equivalent, polar, side, backward and if not W UMa also
        forward radius and assign to component instance.

        :param calculate_equivalent_radius: bool; some application do not require calculation of equivalent radius
        :param components_distance: float; distance of components in SMA unit
        """
        radii = self.calculate_components_radii(components_distance)

        for component, rs in radii.items():
            instance: Star = getattr(self, component)

            for key, value in rs.items():
                setattr(instance, key, value)

            if calculate_equivalent_radius:
                setattr(instance, 'equivalent_radius', self.calculate_equivalent_radius(component)[component])

    def setup_albedos(self):
        """
        Setup of default componet albedo.
        """
        for component, instance in self.components.items():
            instance.albedo = interpolate_albedo(instance.t_eff) if utils.is_empty(instance.albedo) else instance.albedo

    @staticmethod
    def compute_filling_factor(surface_potential, lagrangian_points):
        """
        Compute filling factor of given BinaryStar system.
        Filling factor is computed as::

            (Omega_{inner} - Omega) / (Omega_{inner} - Omega_{outter}),

        where Omega_X denote potential value and `Omega` is potential of given Star.
        Inner and outter are critical inner and outter potentials for given binary star system.

        :param surface_potential: float;
        :param lagrangian_points: List; lagrangian points in `order` (in order to ensure [L3, L1, L2])
        :return: float;
        """
        return (lagrangian_points[1] - surface_potential) / (lagrangian_points[1] - lagrangian_points[2])

    def correct_potentials(self, phases=None, component="all", iterations=2, distances=None):
        """
        Function calculates potential for each phase in phases in such way that conserves
        volume of the component.

        :param phases: numpy.array; if `distances` is not `None`, phases will be not used
        :param component: str; `primary`, `secondary` or `all` (=both)
        :param iterations: int;
        :param distances: numpy.array; if not `None`, corrected potentials will be calculated
                                       for given component distances
        :return: numpy.array;
        """
        if distances is None:
            if phases is None:
                raise ValueError('Either `phases` or components `distances` have to be supplied.')

            data = self.orbit.orbital_motion(phases)
            distances = data[:, 0]
        distances = np.array(distances)
        components = bsutils.component_to_list(component)

        potentials = dict()
        for component in components:
            star = getattr(self, component)
            new_potentials = star.surface_potential * np.ones(distances.shape)

            points_equator, points_meridian = \
                self.generate_equator_and_meridian_points(
                    components_distance=1.0,
                    component=component,
                    surface_potential=star.surface_potential
                )
            volume = utils.calculate_volume_ellipse_approx(points_equator, points_meridian)
            equiv_r_mean = utils.calculate_equiv_radius(volume)

            side_radii = np.empty(distances.shape)
            volume = np.empty(distances.shape)
            for _ in range(iterations):
                for idx, pot in enumerate(new_potentials):
                    radii_args = (star.synchronicity, self.mass_ratio, distances[idx], new_potentials[idx], component)
                    side_radii[idx] = bsradius.calculate_side_radius(*radii_args)

                    points_equator, points_meridian = \
                        self.generate_equator_and_meridian_points(
                            components_distance=distances[idx],
                            component=component,
                            surface_potential=new_potentials[idx]
                        )
                    volume[idx] = utils.calculate_volume_ellipse_approx(points_equator, points_meridian)

                equiv_r = utils.calculate_equiv_radius(volume)
                coeff = equiv_r_mean / equiv_r
                corrected_side_radii = coeff * side_radii

                new_potentials = [
                    bsutils.potential_from_radius(component, corrected_side_radii[idx], const.HALF_PI,
                                                  const.HALF_PI, distance, self.mass_ratio,
                                                  star.synchronicity) for idx, distance in enumerate(distances)
                ]

            potentials[component] = np.array(new_potentials)
        return potentials

    def calculate_equivalent_radius(self, component):
        """
        Function returns equivalent radius of the given component, i.e. radius of the 
        sphere with the same volume as given component.

        :param component: str; `primary`, `secondary` or `all` (=both)
        :return: Dict; {'primary': r_equiv, ...}
        """
        components = bsutils.component_to_list(component)
        r_equiv = dict()
        for component in components:
            star = getattr(self, component)
            points_equator, points_meridian = \
                self.generate_equator_and_meridian_points(
                    components_distance=1.0,
                    component=component,
                    surface_potential=star.surface_potential
                )

            volume = utils.calculate_volume_ellipse_approx(points_equator, points_meridian)
            r_equiv[component] = utils.calculate_equiv_radius(volume)
        return r_equiv

    def calculate_bolometric_luminosity(self, components):
        """
        Calculates bolometric luminosity of given component based on its effective
        temperature and equivalent radius using black body approximation.

        :param components: str; `primary`, `secondary` or `all` (=both)
        :return: Dict; {'primary': L_bol, ...}
        """
        components = bsutils.component_to_list(components)
        r_equiv = {component: getattr(self, component).equivalent_radius for component in components}

        luminosity = dict()
        for component in components:
            star = getattr(self, component)
            luminosity[component] = 4.0 * const.PI * np.power(r_equiv[component] * self.semi_major_axis, 2
                                                              ) * const.STEFAN_BOLTZMAN_CONST * np.power(star.t_eff, 4)

        return luminosity

    def generate_equator_and_meridian_points(self, components_distance, component, surface_potential):
        """
        Function calculates a two arrays of points contouring equator and meridian calculating for the same x-values.

        :param surface_potential: float;
        :param component: str; `primary` or `secondary`
        :param components_distance: float;
        :return: Tuple[numpy.array, numpy.array]; (points on equator, points on meridian)
        """

        forward_radius, x, a = None, None, None
        star = getattr(self, component)
        discretization_factor = star.discretization_factor

        rad_args = (star.synchronicity, self.mass_ratio, components_distance, surface_potential, component)
        backward_radius = bsradius.calculate_backward_radius(*rad_args)

        # generating equidistant angles
        if self.morphology == 'detached':
            num = int(const.PI // discretization_factor)
            theta = np.linspace(discretization_factor, const.PI - discretization_factor, num=num + 1,
                                endpoint=True)

            forward_radius = bsradius.calculate_forward_radius(*rad_args)

            # generating x coordinates for both meridian and equator
            a = 0.5 * (forward_radius + backward_radius)
            c = forward_radius - a
            x = a * up.cos(theta) + c

        elif self.morphology == 'over-contact':
            num = int(const.HALF_PI // discretization_factor)
            theta = np.linspace(const.HALF_PI + discretization_factor, const.PI - discretization_factor, num=num + 1,
                                endpoint=True)

            forward_radius = mesh.calculate_neck_position(self, return_polynomial=False) \
                if component == 'primary' else 1 - mesh.calculate_neck_position(self, return_polynomial=False)

            a = 0.5 * (forward_radius + backward_radius)
            c = forward_radius - a
            x_back = a * up.cos(theta) + c

            x_front = np.linspace(forward_radius, c, num=num + 1, endpoint=True)
            x = np.concatenate((x_front, x_back))

        elif self.morphology in ['semi-detached', 'double-contact']:
            num = int(const.HALF_PI // discretization_factor)
            theta = np.linspace(const.HALF_PI + discretization_factor, const.PI - discretization_factor, num=num,
                                endpoint=True)

            forward_radius = bsradius.calculate_forward_radius(*rad_args)

            a = 0.5 * (forward_radius + backward_radius)
            c = forward_radius - a

            x_front = np.linspace(forward_radius - 0.05 * a, c, num=num + 1, endpoint=True)
            x_back = a * up.cos(theta) + c

            x = np.concatenate((x_front, x_back))

        fn_cylindrical = getattr(model, f"potential_{component}_cylindrical_fn")
        precal_cylindrical = getattr(model, f"pre_calculate_for_potential_value_{component}_cylindrical")
        cylindrical_potential_derivative_fn = getattr(model, f"radial_{component}_potential_derivative_cylindrical")

        phi1, phi2 = const.HALF_PI * np.ones(x.shape), up.zeros(x.shape)
        phi, z = up.concatenate((phi1, phi2)), up.concatenate((x, x))

        args = (phi, z, components_distance, a / 2, precal_cylindrical, fn_cylindrical,
                cylindrical_potential_derivative_fn, surface_potential, self.mass_ratio, star.synchronicity)
        points = mesh.get_surface_points_cylindrical(*args)

        if self.morphology != 'over-contact':
            # add forward and backward points to meridians
            equator_points = np.vstack(([0, 0, forward_radius],
                                        points[:points.shape[0] // 2, :],
                                        [0, 0, -backward_radius]))
            meridian_points = np.vstack(([0, 0, forward_radius],
                                         points[points.shape[0] // 2:, :],
                                         [0, 0, -backward_radius]))

            return equator_points, meridian_points
        else:
            equator_points = np.vstack((points[:points.shape[0] // 2, :], [0, 0, -backward_radius]))
            meridian_points = np.vstack((points[points.shape[0] // 2:, :], [0, 0, -backward_radius]))

            return equator_points, meridian_points

    # light curves *****************************************************************************************************
    def compute_lightcurve(self, **kwargs):
        """
        This function decides which light curve generator function is used.
        Depending on the basic properties of the binary system.

        :param kwargs: Dict; arguments to be passed into light curve generator functions
        :**kwargs options**:
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
            * ** phases ** * - numpy.array
            * ** position_method ** * - method

        :return: Dict; {`passband`: numpy.array, }
        """
        curve_fn = c_router.resolve_curve_method(self, curve='lc')
        return curve_fn(**kwargs)

    def _compute_circular_synchronous_lightcurve(self, **kwargs):
        return lc.compute_circular_synchronous_lightcurve(self, **kwargs)

    def _compute_circular_spotty_asynchronous_lightcurve(self, **kwargs):
        return lc.compute_circular_spotty_asynchronous_lightcurve(self, **kwargs)

    def _compute_circular_pulsating_lightcurve(self, **kwargs):
        return lc.compute_circular_pulsating_lightcurve(self, **kwargs)

    def _compute_eccentric_spotty_lightcurve(self, **kwargs):
        return lc.compute_eccentric_spotty_lightcurve(self, **kwargs)

    def _compute_eccentric_lightcurve(self, **kwargs):
        return lc.compute_eccentric_lightcurve_no_spots(self, **kwargs)

    # radial velocity curves *******************************************************************************************
    def compute_rv(self, **kwargs):
        """
        This function decides which radial velocities generator function is
        used depending on the user-defined method.

        :param kwargs: Dict;
        :**kwargs options**:
            * :method: str; `kinematic` (motion of the centre of mass) or
                            `radiometric` (radiance weighted contribution of each visible element)
            * :position_method: callable; method for obtaining orbital positions
            * :phases: numpy.array; photometric phases

        :return: Dict; {`primary`: numpy.array, `secondary`: numpy.array}
        """
        if kwargs['method'] == 'kinematic':
            return rv.kinematic_radial_velocity(self, **kwargs)
        elif kwargs['method'] == 'radiometric':
            curve_fn = c_router.resolve_curve_method(self, curve='rv')
            kwargs = rv_utils.include_passband_data_to_kwargs(**kwargs)
            return curve_fn(**kwargs)
        else:
            raise ValueError(f"Unknown RV computing method `{kwargs['method']}`.\n"
                             f"List of available methods: [`kinematic`, `radiometric`].")

    def _compute_circular_synchronous_rv_curve(self, **kwargs):
        return rv.compute_circular_synchronous_rv_curve(self, **kwargs)

    def _compute_circular_spotty_asynchronous_rv_curve(self, **kwargs):
        return rv.compute_circular_spotty_asynchronous_rv_curve(self, **kwargs)

    def _compute_circular_pulsating_rv_curve(self, **kwargs):
        return rv.compute_circular_pulsating_rv_curve(self, **kwargs)

    def _compute_eccentric_spotty_rv_curve(self, **kwargs):
        return rv.compute_eccentric_spotty_rv_curve(self, **kwargs)

    def _compute_eccentric_rv_curve_no_spots(self, **kwargs):
        return rv.compute_eccentric_rv_curve_no_spots(self, **kwargs)



import re

from .. params.parameters import deflate_phenomena
from ... import const


def serialize_system_kwargs(**kwargs):
    """
    The function is used to extract the system-related parameters used in synthetic model functions during the fit.
    Those parameters are then returned as `system` component of the input JSON used to initialize the BinarySystem
    instance.

    :param kwargs: Dict[str, float]; model parameters in flat format
    :return: Dict[str, float]; `system` model parameters
    """
    return dict(
        argument_of_periastron=kwargs.get('system@argument_of_periastron', const.HALF_PI),
        gamma=kwargs.get('system@gamma', 0.0),
        period=kwargs["system@period"],
        eccentricity=kwargs.get('system@eccentricity', 0.0),
        primary_minimum_time=0.0,
        **{"inclination": kwargs["system@inclination"]} if kwargs.get("system@inclination") else {},
        **{"semi_major_axis": kwargs["system@semi_major_axis"]} if kwargs.get("system@semi_major_axis") else {},
        **{"mass_ratio": kwargs["system@mass_ratio"]} if kwargs.get("system@mass_ratio") else {},
        **{"asini": kwargs["system@asini"]} if kwargs.get("system@asini") else {},
        **{"additional_light": kwargs["system@additional_light"]} if kwargs.get("system@additional_light") else {},
        **{"phase_shift": kwargs["system@phase_shift"]} if kwargs.get("system@phase_shift") else {}
    )


def _serialize_star_kwargs(component, **kwargs):
    """
    General function for extraction of component-related parameters used in synthetic model functions during the fit.
    Compoenet related parameters are then returned as `primary` or `secondary` component of the input JSON used to
    initialize the BinarySystem instance.

    :param component: str; `primary` or `secondary`
    :param kwargs: Dict; model parameters in flat format
    :return: Dict; component-related model parameters
    """
    prefix = lambda prop: f'{component}@{prop}'
    params_tree = {"spots": dict(), "pulsations": dict()}

    for phenom in params_tree.keys():

        _ = {key: value for key, value in kwargs.items()
             if re.search(rf"^(?=.*\b{phenom[:-1]}\b)(?=.*\b{component}\b).*$", key)}
        _ = deflate_phenomena(_)
        params_tree[phenom] = [{key: val for key, val in value.items() if key not in ['label']} for value in _.values()]

    spots = params_tree['spots'] or dict()
    pulsations = params_tree['pulsations'] or dict()

    return dict(
        surface_potential=kwargs[prefix('surface_potential')],
        synchronicity=kwargs.get(prefix('synchronicity'), 1.0),
        t_eff=kwargs[prefix('t_eff')],
        **{"gravity_darkening": kwargs[prefix("gravity_darkening")]} if kwargs.get(prefix("gravity_darkening")) else {},
        **{"albedo": kwargs[prefix("albedo")]} if kwargs.get(prefix("albedo")) else {},
        **{"metallicity": kwargs[prefix("metallicity")]} if kwargs.get(prefix("metallicity")) else {},
        **{"mass": kwargs[prefix("mass")]} if kwargs.get(prefix("mass")) else {},
        **{"discretization_factor": kwargs[prefix("discretization_factor")]}
        if kwargs.get(prefix("discretization_factor")) else {},
        **{"atmosphere": kwargs[prefix("atmosphere")]} if kwargs.get(prefix("atmosphere")) else {},
        **{"limb_darkening_coefficients": kwargs[prefix("limb_darkening_coefficients")]}
        if kwargs.get(prefix("limb_darkening_coefficients")) else {},
        **{"spots": spots},
        **{"pulsations": pulsations},
    )


def serialize_primary_kwargs(**kwargs):
    """Parameter extractor for the primary component."""
    return _serialize_star_kwargs(component='primary', **kwargs)


def serialize_secondary_kwargs(**kwargs):
    """Parameter extractor for the secondary component."""
    return _serialize_star_kwargs(component='secondary', **kwargs)



from .. models import serializers
from .. tools.utils import time_layer_resolver
from ... base.error import InitialParamsError
from ... binary_system.curves.community import RadialVelocitySystem
from ... binary_system.system import BinarySystem
from ... binary_system.utils import resolve_json_kind


def prepare_central_rv_binary(**kwargs):
    """
    Setup binary system from initial parameter object.

    :param kwargs: Dict; complete set of model parameters in flat format {'parameter@name': value, }
    :return: elisa.binary_system.system.BinarySystem
    """
    return BinarySystem.from_json(kwargs, _verify=False)


def central_rv_synthetic(phases, observer, **kwargs):
    """
    Function returns synthetic RV curve of binary system based on a set of model parameters.

    :param phases: Union[List, numpy.array]; photometric phases in which the curves will be generated
    :param observer: elisa.observer.observer.Observer; observer instance
    :param kwargs: Dict;
    :**kwargs options**:
        * **'system@argument_of_periastron'**: float;
        * **'system@eccentricity'**: float;
        * **'system@inclination'**: float;
        * **'primary@mass'**: float;
        * **'secondary@mass'**: float;
        * **'system@gamma'**: float;
        * **'system@asini'**: float;
        * **'system@mass_ratio'**: float;
        * **'system@primary_minimum_time'**: float;
        * **'system@period'**: float;

    :return: Tuple;
    """

    kwargs.update({
        "primary@surface_potential": 100,
        "secondary@surface_potential": 100,
        "primary@t_eff": 10000.0,
        "secondary@t_eff": 10000.0,
        "primary@metallicity": 10000.0,
        "secondary@metallicity": 10000.0,
    })

    x_data_resolved, kwargs = time_layer_resolver(phases, pop=True, **kwargs)

    system_kwargs = serializers.serialize_system_kwargs(**kwargs)
    primary_kwargs = serializers.serialize_primary_kwargs(**kwargs)
    secondary_kwargs = serializers.serialize_secondary_kwargs(**kwargs)

    json = {
        "primary": dict(**primary_kwargs),
        "secondary": dict(**secondary_kwargs),
        "system": dict(**system_kwargs)
    }

    kind_of = resolve_json_kind(data=json, _sin=True)

    if kind_of in ["std"]:
        observable = prepare_central_rv_binary(**json)
    elif kind_of in ["community"]:
        observable = RadialVelocitySystem(**RadialVelocitySystem.prepare_json(json["system"]))
    else:
        raise InitialParamsError("Initial parameters led to unknown model.")

    observer._system = observable
    observer._system_cls = type(observable)
    _, rv = observer.observe.rv(phases=x_data_resolved, normalize=False)

    return rv



import numpy as np

from elisa.graphic import graphics
from .. import units as u
from ..base.types import FLOAT


class PassbandPlot(object):
    def __init__(self, container):
        self.container = container

    def passband(self):
        xs = np.array(self.container.table.wavelength, dtype=FLOAT)
        ys = np.array(self.container.table.throughput, dtype=FLOAT)
        x_unit = r"Wavelength [$\dot{\AA}$]"
        y_unit = 'Transmissivity [%/100]'

        kwargs = dict(
            xs=xs,
            ys=ys,
            x_unit=x_unit,
            y_unit=y_unit,
            passband=self.container.passband,
        )
        return graphics.passband_plot(**kwargs)


class Plot(object):
    """
    Universal plot interface for Observer class, more detailed documentation for each value of descriptor is
    available in graphics library.

    Available methods::

        `orbit` - plots orbit in orbital plane

    """

    def __init__(self, observer):
        self.observer = observer

    def phase_curve(self, phases=None, fluxes=None, unit=None, legend=True, legend_location=4,
                    return_figure_instance=False):
        """
        Function plots phase curves calculated in Observer class.

        :param phases: numpy.ndarray;
        :param fluxes: Dict; fluxes in each passband
        :param unit: Union[NoneType, astropy.units]; units of flux
        :param legend: bool;
        :param legend_location: int; wrapper for matplotlib `loc` argument for legend location
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        if (phases is None) != (fluxes is None) != (unit is None):
            raise TypeError('You have to supply `phases`, `fluxes` and `unit` variables or none of them.')

        kwargs = {
            "return_figure_instance": return_figure_instance,
            "phases": self.observer.phases if phases is None else phases,
            "fluxes": self.observer.fluxes if fluxes is None else fluxes,
            "unit": self.observer.flux_unit if unit is None else unit,
            "legend": legend,
            "legend_location": legend_location
        }
        if isinstance(unit, type(u.W/u.m**2)) and fluxes is None:
            for _filter, fluxes in kwargs['fluxes'].items():
                kwargs['fluxes'][_filter] = (fluxes * self.observer.flux_unit).to(unit).value

        return graphics.phase_curve(**kwargs)

    def rv_curve(self, phases=None, radial_velocities=None, unit=None, legend=True, legend_location=4,
                 return_figure_instance=False):
        """
        Function plots radial velocity curves calculated in Observer class.

        :param phases: numpy.array;
        :param radial_velocities: Dict;
        :param unit: Union[None, astropy.units.quantity.Quantity]; unit of input 'radial_velocities', if they are not
                     supplied, values calculated in Observer instance are converted to `unit`
        :param legend: bool;
        :param legend_location: int;
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        if (phases is None) != (radial_velocities is None) != (unit is None):
            raise TypeError('You have to supply both `phases` `fluxes`, `radial_velocities` and `None` or none of '
                            'them.')

        kwargs = {
            "return_figure_instance": return_figure_instance,
            "phases": self.observer.phases if phases is None else phases,
            "rvs": self.observer.radial_velocities if radial_velocities is None else radial_velocities,
            "unit": self.observer.rv_unit if unit is None else unit,
            "legend": legend,
            "legend_location": legend_location
        }

        if isinstance(unit, type(u.km/u.s)) and radial_velocities is None:
            for component, rvs in kwargs['rvs'].items():
                kwargs['rvs'][component] = (rvs*self.observer.rv_unit).to(unit).value

        return graphics.rv_curve(**kwargs)

    rv = rv_curve
    lc = phase_curve



import logging
from . import settings

settings.set_up_logging()


# noinspection PyPep8Naming
def getLogger(name, suppress=False):
    if settings.SUPPRESS_LOGGER is not None:
        suppress = settings.SUPPRESS_LOGGER

    return logging.getLogger(name=name) if not suppress else Logger(name)


# noinspection PyPep8Naming
def getPersistentLogger(name):
    return logging.getLogger(name=name)


class Logger(object):
    # noinspection PyUnusedLocal
    def __init__(self, name, *args, **kwargs):
        self.name = name

    def info(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass

    def debug(self, *args, **kwargs):
        pass

    def warning(self, *args, **kwargs):
        pass

    def warn(self, *args, **kwargs):
        pass



from multiprocessing.pool import Pool

from .. import utils
from .. logger import getLogger
from .. import settings

logger = getLogger('observer.mp')


def manage_observations(fn, fn_args, position, **kwargs):
    """
    Function decides whether curve will be calculated using single or multi-process approach.

    :param fn: function used for curve integration
    :param fn_args: Tuple; some of the argument in `fn`
    :param position: List;
    :param kwargs: Dict;
    :return: Dict; calculated curves (in each passbands)
    """
    args = fn_args + (kwargs, )
    if len(position) >= settings.NUMBER_OF_PROCESSES > 1:
        logger.info("starting multiprocessor workers")
        phase_batches = utils.split_to_batches(array=position, n_proc=settings.NUMBER_OF_PROCESSES)
        pool = Pool(processes=settings.NUMBER_OF_PROCESSES)

        result = [pool.apply_async(fn, args[:2] + (batch, ) + args[2:]) for batch in phase_batches]
        pool.close()
        pool.join()
        # this will return output in same order as was given on apply_async init
        result = [r.get() for r in result]
        band_curves = utils.renormalize_async_result(result)
        return band_curves
    else:
        args = args[:2] + (position, ) + args[2:]
        return fn(*args)



import numpy as np
import matplotlib.path as mpltpath

from ... pypex.base import shape
from ... pypex.poly2d.intersection import sat, linter
from ... pypex.poly2d.point import Point
from ... pypex.poly2d.point import is_point_in_polygon
from ... pypex.poly2d.line import Line
from ... pypex.base.conf import ROUND_PRECISION


class Polygon(shape.Shape2D):
    """
    Convex Polygon
    """
    def __init__(self, hull, **kwargs):
        super(Polygon, self).__init__(hull=hull, **kwargs)
        self.sort_clockwise(inplace=True)
        self.mplpath = mpltpath.Path(self.hull)

    def edges(self, as_line=False):
        """
        Provide method to iterate over edges in polygon.
        :param as_line: bool; if True, numpy.array representation of edge is transformed to pypex.poly2d.line.Line
        :return: numpy.array;
        """
        for i in range(-1, len(self.hull)-1, 1):
            edge = np.array([self.hull[i], self.hull[i+1]])
            if as_line:
                edge = Line(edge)
            yield edge

    def intersects(self, poly, in_touch=False, round_tol=ROUND_PRECISION):
        """
        Whether two polygons intersects.

        :param round_tol: int;
        :param poly: pypex.poly2d.polygon.Polygon;
        :param in_touch: bool;
        :return: bool;
        """
        return sat.intersects(self.hull, poly.hull, in_touch, round_tol)

    def intersection(self, poly, round_tol=ROUND_PRECISION):
        """
        Find intersection polygon created by clipping of one polygon by another.

        :param poly: pypex.poly2d.polygon.Polygon
        :param round_tol: int; round precision of decimal points to consider numbers as same
        :return: pypex.poly2d.polygon.Polygon
        """
        in_poly1 = poly.hull[self.mplpath.contains_points(poly.hull)]
        in_poly2 = self.hull[poly.mplpath.contains_points(self.hull)]
        intersection_poly = np.concatenate((in_poly1, in_poly2), axis=0)
        intersection_poly = np.array([Point(*point) for point in intersection_poly])

        _, intersection_segment, intr_ptx, _, msg, _ = linter.intersections(self.hull, poly.hull, in_touch=True)
        points_mask = np.logical_and(msg == b'INTERSECT', intersection_segment)
        intr_ptx = [Point(*point) for point in intr_ptx[points_mask]]
        intersection_poly = np.concatenate((intersection_poly, intr_ptx), axis=0)
        intersection_poly = Point.set(intersection_poly, round_tol=round_tol)
        return Polygon(intersection_poly, _validity=False) if len(intersection_poly) > 2 else None

    def to_array(self):
        return self.hull

    def contains_point(self, point):
        """
        Test wether point lie in polygon.

        :param point: pypex.poly2d.point.Point
        :return:
        """
        return is_point_in_polygon(point, self)

    contains_Point = contains_point

    def surface_area(self):
        """
        Compute surface area of given polygon instance.
        :return: float; surface area of given polygon
        """
        lines = linter.polygon_hull_to_edges(self.hull)
        return 0.5 * np.abs(np.sum(lines[:, 0, 0] * lines[:, 1, 1] - lines[:, 1, 0] * lines[:, 0, 1]))

    def inpolygon(self):
        """
        Find polygon which points are defind as center of each edge of original polygon.
        :return: pypex.poly2d.polygon.Polygon
        """
        _inpolygon = list()
        for edge in self.edges(as_line=True):
            parametrized = edge.parametrized()
            _inpolygon.append(parametrized(0.5))
        return Polygon(_inpolygon)



import numpy as np

from .. import const
from .. umpy import sph_harm
from .. import settings
from .. logger import getLogger

logger = getLogger('pulse.pulsations')


def spherical_harmonics(mode, points, time_exponential, order=None, degree=None):
    """
    Returns spherical harmonics normalized such that its rms = 1.

    :param mode: PulsationMode; mode used to generate sph. harmonics
    :param points: np.array; points in spherical coordinates on which to calculate spherical harmonics value
    :param time_exponential: float; time at which evaluate spherical harmonics
    :param degree: int;
    :param order: int;
    :return: numpy.array; array of spherical harmonics for  given `points`
    """
    l = mode.l if degree is None else degree
    m = mode.m if order is None else order
    return mode.renorm_const * sph_harm(m, l, points[:, 1], points[:, 2]) * time_exponential


def diff_spherical_harmonics_by_phi(mode, harmonics):
    """
    Returns d Y_m^l / d phi

    :param mode: PulsationMode; mode used to generate sph. harmonics
    :param harmonics: List; [Y_l^m, Y_l^m+1]
    :return: numpy.array;
    """
    return (0 + 1j) * mode.m * harmonics[0]


def diff_spherical_harmonics_by_theta(mode, harmonics, phis, thetas):
    """
    Returns d Y_m^l / d theta

    :param mode: PulsationMode; mode used to generate sph. harmonics
    :param harmonics: List; [Y_l^m, Y_l^m+1]
    :param phis: numpy.array;
    :param thetas: numpy.array;
    :return: numpy.array;
    """
    theta_test = np.logical_and(thetas != 0.0, thetas != const.PI)
    derivative = np.zeros(phis.shape, dtype=np.complex128)
    derivative[theta_test] = mode.m * harmonics[0][theta_test] / np.tan(thetas[theta_test]) + \
                             np.sqrt((mode.l - mode.m) * (mode.l + mode.m + 1)) * \
                             np.exp((0 - 1j) * phis[theta_test]) * harmonics[1][theta_test]
    return derivative


def horizontal_displacement_normalization(derivatives, harmonics):
    """
    Normalizes the RMS of horizontal displacement of the given pulsation to 1.

    :param derivatives: numpy.array; 2*N - dY/d_phi, dY/d_theta
    :param harmonics: numpy.array; 2*N - Y^l_m, Y^l_m+1
    :return: float;
    """
    return np.sqrt(np.sum(np.power(np.abs(harmonics[0]), 2)) /
                   np.sum(np.power(np.abs(derivatives[0]), 2) + np.power(np.abs(derivatives[1]), 2)))


def assign_amplitudes(star_container, normalization_constant=1.0):
    """
    Assigns amplitude of radial and horizontal motion for given modes.

    :param normalization_constant: factor to adjust amplitudes, in cas of Binary system it is semi major axis, in case
                                   of single system it should stay 1.0
    :param star_container: StarContainer;
    """
    r_equiv = star_container.equivalent_radius * normalization_constant
    mult = const.G * star_container.mass / r_equiv ** 3
    for mode_index, mode in star_container.pulsations.items():

        # horizontal/radial amplitude (Aerts 2010), p. 198
        if mode.horizontal_to_radial_amplitude_ratio is None:
            amp_ratio = np.sqrt(mode.l * (mode.l + 1)) * mult / mode.angular_frequency ** 2
            mode.horizontal_to_radial_amplitude_ratio = amp_ratio

        amplitude = mode.amplitude / mode.angular_frequency

        mode.radial_amplitude = amplitude / np.sqrt(mode.horizontal_to_radial_amplitude_ratio**2 + 1)
        mode.horizontal_amplitude = mode.horizontal_to_radial_amplitude_ratio * mode.radial_amplitude

        if mode.temperature_amplitude_factor is None:
            if mode.l == 0 or mode.horizontal_to_radial_amplitude_ratio == 0.0:
                raise ValueError('Parameter `temperature_amplitude_factor` needs to be supplied in '
                                 'case of radial modes or in case of modes with radial motion.')
            mode.temperature_amplitude_factor = temp_amplitude(mode) * mode.radial_amplitude / r_equiv

        surf_ampl = mode.horizontal_amplitude / r_equiv
        if surf_ampl > settings.SURFACE_DISPLACEMENT_TOL:
            prec = int(- np.log10(surf_ampl) + 2)
            if not settings.SUPPRESS_WARNINGS:
                logger.warning(f'Relative horizontal surface displacement amplitude '
                               f'({round(surf_ampl, prec)}) for the mode {mode_index} '
                               f'exceeded safe tolerances ({settings.SURFACE_DISPLACEMENT_TOL}) given by the'
                               f' use of linear approximation. This can lead to invalid surface '
                               f'discretization. Use this result with caution.')


def temp_amplitude(mode):
    """
    Returns temperature perturbation amplitude in form of the scalar therm in eq 22 in Townsend 2003::

        delta T = temp_amplitude * (delta r / r) * T;
        temp_amplitude = nabla_ad * (Kl(l+1) - 4 - 1/K)

    where K is our horizontal to radial amplitude ratio.

    :param mode: PulsationMode;
    :return: float;
    """
    return const.IDEAL_ADIABATIC_GRADIENT * (
        mode.horizontal_to_radial_amplitude_ratio * mode.l * (mode.l + 1) - 4 -
        1 / mode.horizontal_to_radial_amplitude_ratio
    )



from ... logger import getLogger
from ... base.surface import temperature as btemperature

logger = getLogger("single_system.surface.temperature")


def build_temperature_distribution(system_container):
    """
    Function calculates temperature distribution on across all faces.

    :param system_container: elisa.single_system.container.SinglePositionContainer;
    :return: elisa.single_system.container.SinglePositionContainer;
    """
    star_container = system_container.star

    logger.debug('Computing effective temperature distribution on the star.')
    star_container.temperatures = \
        btemperature.calculate_effective_temperatures(star_container, star_container.potential_gradient_magnitudes)

    if star_container.has_spots():
        for spot_index, spot in star_container.spots.items():
            logger.debug('Computing temperature distribution of {} spot'.format(spot_index))

            pgms = spot.potential_gradient_magnitudes
            spot.temperatures = \
                spot.temperature_factor * btemperature.calculate_effective_temperatures(star_container, pgms)

    # renormalize_temperatures(star_container)
    return system_container



import numpy as np

from ... import units as u
from ... base.transform import (
    TransformProperties,
    quantity_transform,
    WHEN_FLOAT64
)
from ... base.types import FLOAT, INT


class OrbitProperties(TransformProperties):
    @staticmethod
    def eccentricity(value):
        """
        Transform and validate eccentricity.

        :param value: Union[(numpy.)int, (numpy.)float]
        :return: float
        """
        if not isinstance(value, (int, INT, float, FLOAT)):
            raise TypeError('Input of variable `eccentricity` is not (numpy.)int or (numpy.)float.')
        if value < 0 or value >= 1:
            raise ValueError('Input of variable `eccentricity` is  or it is out of boundaries.')
        return np.float64(value)

    @staticmethod
    def period(value):
        """
        Transform and validate orbital period of binary star system.
        If unit is not specified, default period unit is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultBinarySystemUnits.system.period, WHEN_FLOAT64)

    @staticmethod
    def argument_of_periastron(value):
        """
        If unit is not supplied, value in radians is assumed.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultBinarySystemUnits.system.argument_of_periastron, WHEN_FLOAT64)

    @staticmethod
    def inclination(value):
        """
        If unitless values is supplied, default units suppose to be radians.

        :param value: Union[float, astropy.units.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultBinarySystemUnits.system.inclination, WHEN_FLOAT64)



import numpy as np
from scipy.spatial.qhull import Delaunay

from ... base import spot
from ... base.surface import faces as bfaces
from ... base.types import INT
from ... logger import getLogger
from ... base.surface.faces import set_all_surface_centres, mirror_triangulation

logger = getLogger("single_system.surface.faces")


def build_faces(system_container):
    """
    Function tessellates the stellar surface points into a set of triangles
    covering the star without gaps and overlaps.

    :return: elisa.single_system.container.SinglePositionContainer;
    """
    # build surface if there is no spot specified
    if not system_container.star.spots:
        build_surface_with_no_spots(system_container)
    else:
        build_surface_with_spots(system_container)

    return system_container


def build_surface_with_no_spots(system_container):
    """
    function is calling surface building function for single systems without spots and assigns star's surface to
    star object as its property
    :return:
    """
    star_container = system_container.star
    points_length = star_container.base_symmetry_points_number
    # triangulating only one eighth of the star
    # noinspection PyTypeChecker
    points_to_triangulate = np.append(star_container.symmetry_points(), [[0, 0, 0]], axis=0)
    triangles = single_surface(star_container=star_container, points=points_to_triangulate)
    # removing faces from triangulation, where origin point is included
    # noinspection PyUnresolvedReferences
    triangles = triangles[~(triangles >= points_length).any(1)]
    triangles = triangles[~((points_to_triangulate[triangles] == 0.).all(1)).any(1)]
    # setting number of base symmetry faces
    star_container.base_symmetry_faces_number = INT(np.shape(triangles)[0])
    # let's exploit axial symmetry and fill the rest with the surface of the star
    star_container.faces = mirror_triangulation(triangles, star_container.inverse_point_symmetry_matrix)

    base_face_symmetry_vector = np.arange(star_container.base_symmetry_faces_number)
    star_container.face_symmetry_vector = np.concatenate([base_face_symmetry_vector for _ in range(8)])


def single_surface(star_container=None, points=None):
    """
    calculates triangulation of given set of points, if points are not given, star surface points are used. Returns
    set of triple indices of surface pints that make up given triangle

    :param star_container: StarContainer;
    :param points: np.array:

    ::

        numpy.array([[x1 y1 z1],
                     [x2 y2 z2],
                     ...
                    [xN yN zN]])

    :return: np.array():

    ::

        numpy.array([[point_index1 point_index2 point_index3],
                     [...],
                      ...
                     [...]])
    """
    if points is None:
        points = star_container.points
    triangulation = Delaunay(points)
    triangles_indices = triangulation.convex_hull
    return triangles_indices


def build_surface_with_spots(system_container):
    """
    function for triangulation of surface with spots

    :return:
    """
    star_container = system_container.star
    points, vertices_map = star_container.get_flatten_points_map()
    faces = single_surface(points=points)
    model, spot_candidates = bfaces.initialize_model_container(vertices_map)
    model = bfaces.split_spots_and_component_faces(
        star_container, points, faces, model, spot_candidates, vertices_map, component_com=0.0
    )

    spot.remove_overlaped_spots_by_vertex_map(star_container, vertices_map)
    spot.remap_surface_elements(star_container, model, points)

    return system_container


def compute_all_surface_areas(system_container):
    """
    Compute surface areas of all faces (spots included).

    :param system_container: elisa.single_system.container.SinglePositionContainer; instance
    :return: system; elisa.single_system.container.SinglePositionContainer; instance
    """
    star_container = system_container.star
    logger.debug(f'computing surface areas of component: '
                 f'{star_container} / name: {star_container.name}')
    star_container.calculate_all_areas()

    return system_container


def build_faces_orientation(system_container):
    """
    Compute face orientation (normals) for each face.

    :param system_container: elisa.single_system.container.SinglePositionContainer;
    :return: elisa.single_system.container.SinglePositionContainer;
    """
    com_x = 0.0

    star = system_container.star
    set_all_surface_centres(star)
    set_all_normals(star, com=com_x)

    return system_container


def set_all_normals(star_container, com):
    """
    Function calculates normals for each face of given body (including spots) and assign it to object.

    :param star_container: instance of container to set normals on;
    :param com: numpy.array;
    :param star_container: instance of container to set normals on;
    """
    points, faces, cntrs = star_container.points, star_container.faces, star_container.face_centres
    star_container.normals = bfaces.calculate_normals(points, faces, cntrs, com)

    if star_container.has_spots():
        for spot_index in star_container.spots:
            star_container.spots[spot_index].normals = \
                bfaces.calculate_normals(star_container.spots[spot_index].points,
                                         star_container.spots[spot_index].faces,
                                         star_container.spots[spot_index].face_centres,
                                         com)
    return star_container


# noinspection PyUnreachableCode
def build_velocities(system):
    """
    Function calculates velocity vector for each face relative to the observer.

    :param system: elisa.single_system.container.SingleSystemContainer
    :return: elisa.single_system.container.SingleSystemContainer
    """
    star = system.star
    omega = np.array([0, 0, system.angular_velocity])

    # rotational velocity
    p_velocities = np.cross(star.points, omega, axisa=1)
    star.velocities = np.mean(p_velocities[star.faces], axis=1)

    if star.has_spots():
        for _spot in star.spots.values():
            p_velocities = np.cross(_spot.points, omega, axisa=1)
            _spot.velocities = np.mean(p_velocities[_spot.faces], axis=1)

    return system



import os
import numpy as np
import elisa.umpy as up

from .. import model
from .. radius import calculate_radius
from ... logger import getLogger
from ... base.spot import incorporate_spots_mesh
from ... base.surface.mesh import correct_component_mesh
from ... base.error import MaxIterationError
from ... import settings
from ... import (
    opt,
    const,
    utils
)


logger = getLogger("single_system.surface.mesh")
CORRECTION_FACTORS = np.load(settings.PATH_TO_SINGLE_CORRECTIONS, allow_pickle=False)


def build_mesh(system):
    """
    Build surface point mesh including spots.

    :param system: elisa.single_system.contaier.PositionContainer; instance
    :return: elisa.single_system.container.SinglePositionContainer; instance
    """
    a, c, d = mesh(system_container=system, symmetry_output=True)

    system.star.points = a
    system.star.base_symmetry_points_number = c
    system.star.inverse_point_symmetry_matrix = d

    add_spots_to_mesh(system)
    return system


def mesh(system_container, symmetry_output=False):
    """
    Function for creating surface mesh of single star system.

    :return: numpy.array;

    ::

            numpy.array([[x1 y1 z1],
                          [x2 y2 z2],
                            ...
                          [xN yN zN]]) - array of surface points if symmetry_output = False, else:
            numpy.array([[x1 y1 z1],
                          [x2 y2 z2],
                            ...
                          [xN yN zN]]) - array of surface points,
            numpy.float - number of points included in symmetrical one eighth of surface,
            numpy.array([octants[indexes_of_remapped_points_in_octants]) - matrix of eight sub matrices that mapps
                                                                           base symmetry octant to all others
                                                                           octants

    """
    star_container = getattr(system_container, 'star')
    discretization_factor = star_container.discretization_factor
    if discretization_factor > const.HALF_PI:
        raise ValueError("Invalid value of alpha parameter. Use value less than 90.")

    potential_fn = model.potential_fn
    precalc_fn = model.pre_calculate_for_potential_value
    potential_derivative_fn = model.radial_potential_derivative

    characteristic_distance = discretization_factor * star_container.polar_radius

    # calculating equatorial part
    equator_points = calculate_equator_points(characteristic_distance, star_container.equatorial_radius)
    x_eq, y_eq, z_eq = equator_points[:, 0], equator_points[:, 1], equator_points[:, 2]

    # axial symmetry, therefore calculating latitudes
    thetas = pre_calc_latitudes(discretization_factor, star_container.polar_radius, star_container.equatorial_radius)
    thetas_meridian = pre_calc_latitudes(const.SEAM_CONST*discretization_factor, star_container.polar_radius,
                                         star_container.equatorial_radius)

    x0 = 0.5 * (star_container.equatorial_radius + star_container.polar_radius)
    args = thetas, x0, precalc_fn, potential_fn, potential_derivative_fn, star_container.surface_potential, \
        star_container.mass, system_container.angular_velocity
    args_meridian = thetas_meridian, x0, precalc_fn, potential_fn, potential_derivative_fn, \
                    star_container.surface_potential, star_container.mass, system_container.angular_velocity

    radius = get_surface_points_radii(*args)
    radius_meridian = get_surface_points_radii(*args_meridian)

    # converting this eighth of surface to cartesian coordinates
    quarter_points = calculate_points_on_quarter_surface(radius, thetas, characteristic_distance)
    x_q, y_q, z_q = quarter_points[:, 0], quarter_points[:, 1], quarter_points[:, 2]
    meridian_points = calculate_points_on_meridian(radius_meridian, thetas_meridian)
    x_mer, y_mer, z_mer = meridian_points[:, 0], meridian_points[:, 1], meridian_points[:, 2],

    x = np.concatenate((np.array([0]), x_mer, x_eq, x_q, -y_mer, -y_eq, -y_q, -x_mer, -x_eq, -x_q, y_mer, y_eq,
                        y_q, np.array([0]), x_mer, x_q, -y_mer, -y_q, -x_mer, -x_q, y_mer, y_q))
    y = np.concatenate((np.array([0]), y_mer, y_eq, y_q, x_mer, x_eq, x_q, -y_mer, -y_eq, -y_q, -x_mer, -x_eq,
                        -x_q, np.array([0]), y_mer, y_q, x_mer, x_q, -y_mer, -y_q, -x_mer, -x_q))
    z = np.concatenate((np.array([star_container.polar_radius]), z_mer, z_eq, z_q, z_mer, z_eq, z_q, z_mer, z_eq,
                        z_q, z_mer, z_eq, z_q, np.array([-star_container.polar_radius]), -z_mer, -z_q, -z_mer, -z_q,
                        -z_mer, -z_q, -z_mer, -z_q))

    if symmetry_output:
        quarter_equator_length = len(x_eq)
        meridian_length = len(x_mer)
        quarter_length = len(x_q)
        base_symmetry_points_number = 1 + meridian_length + quarter_equator_length + quarter_length + meridian_length

        south_pole_index = 4 * (base_symmetry_points_number - meridian_length) - 3
        reduced_bspn = base_symmetry_points_number - meridian_length  # auxiliary variable1
        reduced_bspn2 = base_symmetry_points_number - quarter_equator_length
        inverse_symmetry_matrix = \
            np.array([
                np.arange(base_symmetry_points_number + 1),  # 1st quadrant (north hem)
                # 2nd quadrant (north hem)
                np.concatenate(([0], np.arange(reduced_bspn, 2 * base_symmetry_points_number - meridian_length))),
                # 3rd quadrant (north hem)
                np.concatenate(([0], np.arange(2 * reduced_bspn - 1, 3 * reduced_bspn + meridian_length - 1))),
                # 4th quadrant (north hem)
                np.concatenate(([0], np.arange(3 * reduced_bspn - 2, 4 * reduced_bspn - 3),
                                np.arange(1, meridian_length + 2))),
                # 1st quadrant (south hemisphere)
                np.concatenate((np.arange(south_pole_index, meridian_length + 1 + south_pole_index),
                                np.arange(1 + meridian_length, 1 + meridian_length + quarter_equator_length),
                                np.arange(meridian_length + 1 + south_pole_index,
                                          base_symmetry_points_number - quarter_equator_length + south_pole_index),
                                [base_symmetry_points_number])),
                # 2nd quadrant (south hem)
                np.concatenate(([south_pole_index],
                                np.arange(reduced_bspn2 - meridian_length + south_pole_index,
                                          reduced_bspn2 + south_pole_index),
                                np.arange(base_symmetry_points_number,
                                          base_symmetry_points_number + quarter_equator_length),
                                np.arange(reduced_bspn2 + south_pole_index,
                                          2 * reduced_bspn2 - meridian_length - 1 +
                                          south_pole_index),
                                [2 * base_symmetry_points_number - meridian_length - 1])),
                # 3rd quadrant (south hem)
                np.concatenate(([south_pole_index],
                                np.arange(2 * reduced_bspn2 - 2 * meridian_length - 1 + south_pole_index,
                                          2 * reduced_bspn2 - meridian_length - 1 + south_pole_index),
                                np.arange(2 * base_symmetry_points_number - meridian_length - 1,
                                          2 * base_symmetry_points_number - meridian_length + quarter_equator_length
                                          - 1),
                                np.arange(2 * reduced_bspn2 - meridian_length - 1 + south_pole_index,
                                          3 * reduced_bspn2 - 2 * meridian_length - 2 + south_pole_index),
                                [3 * reduced_bspn + meridian_length - 2])),
                # 4th quadrant (south hem)
                np.concatenate(([south_pole_index],
                                np.arange(3 * reduced_bspn2 - 3 * meridian_length - 2 + south_pole_index,
                                          3 * reduced_bspn2 - 2 * meridian_length - 2 + south_pole_index),
                                np.arange(3 * reduced_bspn + meridian_length - 2,
                                          3 * reduced_bspn + meridian_length - 2 +
                                          quarter_equator_length),
                                np.arange(3 * reduced_bspn2 - 2 * meridian_length - 2 + south_pole_index, len(x)),
                                np.arange(1 + south_pole_index, meridian_length + south_pole_index + 1),
                                [1 + meridian_length]
                                ))
            ])

        return np.column_stack((x, y, z)), base_symmetry_points_number + 1, inverse_symmetry_matrix
    else:
        return np.column_stack((x, y, z))


def pre_calc_latitudes(alpha, polar_radius, equatorial_radius):
    """
    Function pre-calculates latitudes of stellar surface with exception of pole and equator.

    :param equatorial_radius: float;
    :param polar_radius: float;
    :param alpha: float; angular distance of points
    :return: numpy.array; latitudes for mesh
    """
    # alpha_corr = const.POINT_ROW_SEPARATION_FACTOR * alpha
    alpha_corr = alpha
    num = int(const.HALF_PI // alpha_corr)
    thetas = np.linspace(0, const.HALF_PI, num=num, endpoint=True)[1:-1]
    # solving non uniform sampling along theta coordinates for squashed stars
    auto_test = settings.MESH_GENERATOR == 'auto' and \
                (equatorial_radius - polar_radius) / polar_radius > settings.DEFORMATION_TOL
    if auto_test or settings.MESH_GENERATOR == 'improved_trapezoidal':
        thetas += up.arctan((equatorial_radius - polar_radius) * up.tan(thetas) /
                            (polar_radius + equatorial_radius * up.tan(thetas)**2))
    return thetas


def get_surface_points_radii(*args):
    """
    Function solves radius for given latitudes that are passed in `args`.
    Function to solve is specified as last parameter in `args` Tuple.

    :param args: Tuple;

    ::

        Tuple[
                theta: numpy.array,
                x0: float,
                precalc: callable,
                fn: callable,
                derivative_fn: callable
              ]

    :return: numpy.array;
    """
    theta, x0, precalc_fn, potential_fn, potential_derivative_fn, surface_potential, mass, angular_velocity = args
    precalc_vals = precalc_fn(*(mass, angular_velocity, theta,), return_as_tuple=True)
    x0 = x0 * np.ones(theta.shape)
    radius = opt.newton.newton(potential_fn, x0, fprime=potential_derivative_fn,
                               maxiter=settings.MAX_SOLVER_ITERS, args=(precalc_vals, surface_potential), rtol=1e-10)
    return radius


def calculate_points_on_quarter_surface(radius, thetas, characteristic_distance):
    """
    Function returns cartesian coordinates for points on the quarter of the surface.

    :param radius: numpy.array;
    :param thetas: numpy.array;
    :param characteristic_distance: float; mean distance between points
    :return: numpy.array; N * 3 array of x, y, z coordinates
    """
    r_q, phi_q, theta_q = [], [], []
    for ii, theta in enumerate(thetas):
        num = int(const.HALF_PI * radius[ii] * np.sin(theta) / characteristic_distance)
        alpha = (const.HALF_PI / num)
        num -= 1 if num > 0 else num
        r_q.append(radius[ii] * np.ones(num))
        theta_q.append(theta * np.ones(num))
        phi_q.append(np.linspace(alpha, const.HALF_PI-alpha, num=num, endpoint=True))
    r_q = np.concatenate(r_q)
    theta_q = np.concatenate(theta_q)
    phi_q = np.concatenate(phi_q)
    return utils.spherical_to_cartesian(np.column_stack((r_q, phi_q, theta_q)))


def calculate_points_on_meridian(radius, thetas):
    """
    Function returns cartesian coordinates for points on the surface meridian.

    :param radius: numpy.array;
    :param thetas: numpy.array;
    :return: numpy.array; N * 3 array of x, y, z coordinates
    """
    phi = 0.0 * np.ones(radius.shape)
    return utils.spherical_to_cartesian(np.column_stack((radius, phi, thetas)))


def calculate_equator_points(characteristic_distance, equatorial_radius):
    """
    Function returns cartesian coordinates for points on the equator.

    :param characteristic_distance: float; number of points on the quarter of the equator
    :param equatorial_radius: float;
    :return: numpy.array; N * 3 array of x, y, z coordinates
    """
    num = int(const.HALF_PI * equatorial_radius / (const.SEAM_CONST * characteristic_distance))
    radii = equatorial_radius * np.ones(num)
    thetas = const.HALF_PI * np.ones(num)
    phis = np.linspace(0, const.HALF_PI, num=num, endpoint=False)
    return utils.spherical_to_cartesian(np.column_stack((radii, phis, thetas)))


def mesh_spots(system_container):
    """
    Compute points of each spots and assigns values to spot container instance.
    """

    logger.info("evaluating spots")
    star_container = system_container.star
    if not star_container.has_spots():
        logger.info("no spots to evaluate")
        return

    potential_fn = model.potential_fn
    precalc_fn = model.pre_calculate_for_potential_value
    potential_derivative_fn = model.radial_potential_derivative

    # iterate over spots
    for spot_index, spot_instance in list(star_container.spots.items()):
        # lon -> phi, lat -> theta
        lon, lat = spot_instance.longitude, spot_instance.latitude

        alpha, spot_radius = spot_instance.discretization_factor, spot_instance.angular_radius

        # initial containers for current spot
        boundary_points, spot_points = [], []

        # initial radial vector
        radial_vector = np.array([1.0, lon, lat])  # unit radial vector to the center of current spot
        center_vector = utils.spherical_to_cartesian([1.0, lon, lat])

        args = (radial_vector[2],)

        solution = calculate_radius(star_container.mass, system_container.angular_velocity,
                                    star_container.surface_potential, *args)

        if solution > star_container.equatorial_radius or solution < star_container.polar_radius:
            # in case of spots, each point should be usefull, otherwise remove spot from
            # component spot list and skip current spot computation
            logger.info(f"center of spot {spot_instance.kwargs_serializer()} doesn't satisfy "
                        f"reasonable conditions and entire spot will be omitted")

            star_container.remove_spot(spot_index=spot_index)
            continue

        spot_center_r = solution
        spot_center = utils.spherical_to_cartesian([spot_center_r, lon, lat])

        # compute euclidean distance of two points on spot (x0)
        # we have to obtain distance between center and 1st point in 1st ring of spot
        args = (lat + alpha,)
        solution = calculate_radius(star_container.mass, system_container.angular_velocity,
                                    star_container.surface_potential, *args)

        if solution > star_container.equatorial_radius or solution < star_container.polar_radius:
            # in case of spots, each point should be usefull, otherwise remove spot from
            # component spot list and skip current spot computation
            logger.info(f"first ring of spot {spot_instance.kwargs_serializer()} doesn't satisfy "
                        f"reasonable conditions and entire spot will be omitted")
            star_container.remove_spot(spot_index=spot_index)
            continue

        x0 = np.sqrt(spot_center_r ** 2 + solution ** 2 - (2.0 * spot_center_r * solution * np.cos(alpha)))

        # number of points in latitudal direction
        num_radial = int(np.round(spot_radius / alpha)) + 1
        thetas = np.linspace(lat, lat + spot_radius, num=num_radial, endpoint=True)

        num_azimuthal = [1 if i == 0 else int(i * 2.0 * np.pi * x0 // x0) for i in range(0, len(thetas))]
        deltas = [np.linspace(0., const.FULL_ARC, num=num, endpoint=False) for num in num_azimuthal]

        spot_phi, spot_theta = [], []
        for theta_index, theta in enumerate(thetas):
            # first point of n-th ring of spot (counting start from center)
            default_spherical_vector = [1.0, lon % const.FULL_ARC, theta]

            for delta_index, delta in enumerate(deltas[theta_index]):
                # rotating default spherical vector around spot center vector and thus generating concentric
                # circle of points around centre of spot
                delta_vector = utils.arbitrary_rotation(theta=delta, omega=center_vector,
                                                        vector=utils.spherical_to_cartesian(default_spherical_vector),
                                                        degrees=False,
                                                        omega_normalized=True)

                spherical_delta_vector = utils.cartesian_to_spherical(delta_vector)

                spot_phi.append(spherical_delta_vector[1])
                spot_theta.append(spherical_delta_vector[2])

        spot_phi, spot_theta = np.array(spot_phi), np.array(spot_theta)
        args = spot_theta, spot_center_r, precalc_fn, potential_fn, potential_derivative_fn, \
            star_container.surface_potential, star_container.mass, system_container.angular_velocity
        try:
            spot_points_radii = get_surface_points_radii(*args)
        except MaxIterationError:
            if not settings.SUPPRESS_WARNINGS:
                logger.warning(f"at least 1 point of spot {spot_instance.kwargs_serializer()} "
                               f"doesn't satisfy reasonable conditions and entire spot will be omitted")
            star_container.remove_spot(spot_index=spot_index)
            continue

        spot_points = utils.spherical_to_cartesian(np.column_stack((spot_points_radii, spot_phi, spot_theta)))

        spot_instance.points = np.array(spot_points)
        spot_instance.boundary = np.array(boundary_points)
        spot_instance.boundary_center = spot_points[0]
        spot_instance.center = np.array(spot_center)


def add_spots_to_mesh(system):
    """
    Function implements surface points into clean mesh and removes stellar
    points and other spot points under the given spot if such overlapped spots exists.
    """
    mesh_spots(system)
    incorporate_spots_mesh(system.star, component_com=0.0)


def correct_mesh(system):
    """
    Correcting the underestimation of the surface due to the discretization.

    :param system: elisa.single_system.container.SystemContainer;
    :return: elisa.single_system.container.SystemContainer;
    """
    star = getattr(system, 'star')
    correct_component_mesh(star, com=0.0, correction_factors=CORRECTION_FACTORS)

    return system



import numpy as np

from ... import units as u
from ... base.transform import (
    TransformProperties,
    WHEN_FLOAT64,
    WHEN_ARRAY
)
from ... utils import is_empty
from ... base.types import FLOAT


def array_transform(value, when_array):
    """
    Check whether `value` is array-like and then transforms it to numpy.array.

    :param value: Union[(numpy.)array, list, tuple];
    :param when_array: Tuple(Types);
    :return: numpy.array;
    """
    if isinstance(value, when_array):
        return np.array(value, dtype=FLOAT)
    elif not is_empty(value):
        raise TypeError('Input of variable is not array-like.')


def unit_check(value, base_units):
    """
    Checking if the supplied unit is equivalent to the base unit for given variable.

    :param value: Union[str, astropy.unit.Unit]; unit which compatibility will be checked
    :param base_units: List[astropy.unit.Unit]; base units of given parameter
    :return: astropy.unit.Unit;
    """
    if value is None or value.to_string() == '':
        value = u.dimensionless_unscaled

    if not value.is_equivalent(base_units):
        raise ValueError(f'Input {value} is not NoneType or `astropy.Unit` not convertible into desired base units.')

    return value


class DatasetProperties(TransformProperties):
    """
    Transforming various input time series x,y and y_err to numpy.array format.
    """
    @staticmethod
    def x_data(value):
        return array_transform(value, WHEN_ARRAY)

    @staticmethod
    def y_data(value):
        return array_transform(value, WHEN_ARRAY)

    @staticmethod
    def y_err(value):
        return array_transform(value, WHEN_ARRAY)


class RVDataProperties(DatasetProperties):
    """
    Making sure that time and RV units are convertible to the ELISa's base units.
    """
    @staticmethod
    def x_unit(value):
        return unit_check(value, (u.dimensionless_unscaled, u.TIME_UNIT))

    @staticmethod
    def y_unit(value):
        return unit_check(value, (u.VELOCITY_UNIT,))


class LCDataProperties(DatasetProperties):
    """
    Making sure that time and LC units are convertible to the ELISa's base units.
    """
    @staticmethod
    def x_unit(value):
        return unit_check(value, (u.dimensionless_unscaled, u.TIME_UNIT))

    @staticmethod
    def y_unit(value):
        return unit_check(value, (u.dimensionless_unscaled, u.mag))

    @staticmethod
    def zero_magnitude(value):
        if isinstance(value, u.Quantity):
            value.is_equivalent(u.mag)
            value = np.float64(value.to(u.mag))
        elif isinstance(value, WHEN_FLOAT64):
            value = np.float64(value)
        else:
            raise TypeError('Input of variable is not (numpy.)int or (numpy.)float '
                            'nor astropy.unit.quantity.Quantity instance.')
        return value



import numpy as np


def projection(vector_a, vector_d):
    """
    Projection vector ``p`` (scalar) of vector ``A`` in direction defined by unit vector ``E`` is computed as dot
    product of following::

        (1)   p = (A . E) * E = |E| * |A| * cos(t) * E

    where ``t`` is angle between ``A`` and ``E``.
    Instead of unit vector ``E`` we can use arbitrary vector in direction of ``E``, e.g. ``B`` and compute this
    unit vector from vector ``B`` as following::

        E = B / |B|

    After substitution, we get::

        p = (A . E) * E = ((A . B) / |B|) * (B / |B|) = ((A . B) / (|B|^2)) * B = ((A . B) / (B . B)) * B

    :param vector_a: numpy.array; vector to project
    :param vector_d: numpy.array; direction vector
    :return: numpy.array
    """

    vector_d, vector_a = np.array(vector_d), np.array(vector_a)
    return (np.dot(vector_a, vector_d) / np.dot(vector_d, vector_d)) * vector_d


def cartesian_to_vectors_defined(tn, nn, vector):
    """
    Transform coordinate from standard cartesian 2d defined by unit vector e1 and e2 to coordinate system defined
    by perpendicular vectors 'tn' and 'nn' (tangential and normal vector).
    'tn' represents like new 'x' axis::

        |c1|   | e1 . t    e2 . n |   | v1 |
        |  | = |                  | x |    |
        |c2|   | e1 . n    e2 . t |   | v2 |

    :param tn: numpy.array
    :param nn: numpy.array
    :param vector: numpy.array
    :return: numpy.array
    """

    m = [[np.dot(np.array([1.0, 0.0]), tn), np.dot(np.array([0.0, 1.0]), tn)],
         [np.dot(np.array([1.0, 0.0]), nn), np.dot(np.array([0.0, 1.0]), nn)]]
    return np.dot(m, vector)



import datetime

import numpy as np

from ... import units as u
from ... atm import atm_file_prefix_to_quantity_list
from ... import settings
from ... utils import jd_from_datetime

PARAM_PARSER = '@'
NUISANCE_PARSER = 'nuisance'

TEMPERATURES = atm_file_prefix_to_quantity_list("temperature", settings.ATM_ATLAS)
METALLICITY = atm_file_prefix_to_quantity_list("metallicity", settings.ATM_ATLAS)


COMPOSITE_FLAT_PARAMS = [
    'spot',
    'pulsation'
]

DEFAULT_NORMALIZATION_SPOT = {
    "longitude": (0, 360),
    "latitude": (0, 180),
    "angular_radius": (0, 90),
    "temperature_factor": (0.1, 3),
}

DEFAULT_NORMALIZATION_NUISANCE = {
    "ln_f": (-20, -10),
}

DEFAULT_NORMALIZATION_PULSATION = {
    "l": (0, 10),
    "m": (-10, 10),
    "amplitude": (0, 5000),
    "frequency": (0.01, 40),
    "start_phase": (0, 360),
    "mode_axis_theta": (0, 180),
    "mode_axis_phi": (0, 360)
}

DEFAULT_NORMALIZATION_STAR = {
    "mass": (0.1, 50),
    "t_eff": (np.min(TEMPERATURES), np.max(TEMPERATURES)),
    "metallicity": (np.min(METALLICITY), np.max(METALLICITY)),
    "surface_potential": (2.0, 50.0),
    "albedo": (0, 1),
    "gravity_darkening": (0, 1),
    "synchronicity": (0.01, 10),
}

CURRENT_TIME_JD = jd_from_datetime(datetime.datetime.now())

DEFAULT_NORMALIZATION_SYSTEM = {
    "inclination": (0, 180),
    "eccentricity": (0, 0.9999),
    "argument_of_periastron": (0, 360),
    "gamma": (0, 1e6),
    "mass_ratio": (1e-6, 2),
    "semi_major_axis": (0.01, 100),
    "asini": (0.0001, 100),
    "period": (0.001, 100),
    "additional_light": (0, 1.0),
    "phase_shift": (-0.8, 0.8),
    "primary_minimum_time": (CURRENT_TIME_JD - 365.0, CURRENT_TIME_JD),
}

SPOTS_PARAMETERS = ['longitude', 'latitude', 'angular_radius', 'temperature_factor', 'angular_radius']
PULSATIONS_PARAMETERS = ['l', 'm', 'amplitude', 'frequency', 'start_phase', 'mode_axis_phi', 'mode_axis_theta']

DEFAULT_FLOAT_ANGULAR_UNIT = u.deg
DEFAULT_FLOAT_MASS_UNIT = u.solMass

DEFAULT_FLOAT_UNITS = {
    'inclination': u.DefaultSystemInputUnits.inclination,
    'eccentricity': u.DefaultBinarySystemInputUnits.system.eccentricity,
    'argument_of_periastron': u.DefaultBinarySystemInputUnits.system.argument_of_periastron,
    'gamma': u.DefaultSystemInputUnits.gamma,
    'mass': u.DefaultStarInputUnits.mass,
    't_eff': u.DefaultStarInputUnits.t_eff,
    'metallicity': u.DefaultStarInputUnits.metallicity,
    'surface_potential': u.DefaultBinarySystemInputUnits.component.surface_potential,
    'albedo': u.DefaultStarUnits.albedo,
    'gravity_darkening': u.DefaultStarInputUnits.gravity_darkening,
    'synchronicity': u.DefaultBinarySystemInputUnits.component.synchronicity,
    'mass_ratio': u.DefaultBinarySystemInputUnits.system.mass_ratio,
    'semi_major_axis': u.DefaultBinarySystemInputUnits.system.semi_major_axis,
    'asini': u.solRad,
    'period': u.DefaultBinarySystemInputUnits.system.period,
    'primary_minimum_time': u.DefaultBinarySystemInputUnits.system.primary_minimum_time,
    'additional_light': u.DefaultSystemInputUnits.additional_light,
    'phase_shift': u.DefaultSystemInputUnits.phase_shift,
    # SPOTS
    'latitude': u.DefaultSpotInputUnits.latitude,
    'longitude': u.DefaultSpotInputUnits.longitude,
    'angular_radius': u.DefaultSpotInputUnits.angular_radius,
    'temperature_factor': u.DefaultSpotInputUnits.temperature_factor,
    # PULSATIONS
    'l': u.DefaultPulsationsInputUnits.l,
    'm': u.DefaultPulsationsInputUnits.m,
    'amplitude': u.DefaultPulsationsInputUnits.amplitude,
    'frequency': u.DefaultPulsationsInputUnits.frequency,
    'start_phase': u.DefaultPulsationsInputUnits.start_phase,
    'mode_axis_theta': u.DefaultPulsationsInputUnits.mode_axis_theta,
    'mode_axis_phi': u.DefaultPulsationsInputUnits.mode_axis_phi,
    # NUISANCE
    'ln_f': None
}

PARAMS_KEY_TEX_MAP = {
    'system@argument_of_periastron': '$\\omega$',
    'system@inclination': '$i$',
    'system@eccentricity': '$e$',
    'system@gamma': '$\\gamma$',
    'system@mass_ratio': '$q$',
    'system@semi_major_axis': '$a$',
    'primary@mass': '$M_1$',
    'primary@t_eff': '$T_1^{eff}$',
    'primary@surface_potential': '$\\Omega_1$',
    'primary@gravity_darkening': '$\\beta_1$',
    'primary@albedo': '$A_1$',
    'primary@metallicity': '$M/H_1$',
    'primary@synchronicity': '$F_1$',

    'secondary@mass': '$M_2$',
    'secondary@t_eff': '$T_2^{eff}$',
    'secondary@surface_potential': '$\\Omega_2$',
    'secondary@gravity_darkening': '$\\beta_2$',
    'secondary@albedo': '$A_2$',
    'secondary@metallicity': '$M/H_2$',
    'secondary@synchronicity': '$F_2$',

    'system@asini': 'a$sin$(i)',
    'system@period': '$period$',
    'system@primary_minimum_time': '$T_0$',
    'system@additional_light': '$l_{add}$',
    'system@phase_shift': 'phase shift$',
    # SPOTS
    'longitude': '$\\varphi$',
    'latitude': '$\\vartheta$',
    'angular_radius': '$R_{spot}$',
    'temperature_factor': '$T_{spot}/T_{eff}$',
    # PULSATIONS
    'l': '$\\ell$',
    'm': '$m$',
    'amplitude': '$A$',
    'frequency': '$f$',
    'start_phase': '$\\Phi_0$',
    'mode_axis_phi': '$\\phi_{mode}$',
    'mode_axis_theta': '$\\theta_{mode}$',
    # NUISANCE
    'nuisance@ln_f': "$ln(f)$"
}



import numpy as np

from ... import const as c
from ... logger import getLogger
from ... base.surface import gravity as bgravity

logger = getLogger("single_system.surface.gravity")


def build_surface_gravity(system_container):
    """
    Function calculates gravity potential gradient magnitude (surface gravity) for each face.
    """
    star_container = system_container.star

    polar_gravity = np.power(10, star_container.polar_log_g)

    star_container.polar_potential_gradient_magnitude = polar_gravity

    logger.debug('computing potential gradient magnitudes distribution on a star')
    points, faces = bgravity.eval_args_for_magnitude_gradient(star_container)

    g_acc_vector = calculate_potential_gradient(points, system_container.angular_velocity, star_container.mass)
    gravity = np.mean(np.linalg.norm(g_acc_vector, axis=1)[faces], axis=1)

    setattr(star_container, 'potential_gradient_magnitudes', star_container.mirror_face_values(gravity)) \
        if star_container.symmetry_test() else setattr(star_container, 'potential_gradient_magnitudes', gravity)
    setattr(star_container, 'log_g', np.log10(star_container.potential_gradient_magnitudes))

    if star_container.has_spots():
        g_acc_vector_spot = dict()
        for spot_index, spot in star_container.spots.items():
            logger.debug(f'calculating surface SI unit gravity of {spot_index} spot')
            logger.debug(f'calculating distribution of potential gradient '
                         f'magnitudes of spot index: {spot_index} component')

            g_acc_vector_spot.update(
                {spot_index:
                     calculate_potential_gradient(spot.points, system_container.angular_velocity, star_container.mass)}
            )

            setattr(spot, 'potential_gradient_magnitudes',
                    np.mean(np.linalg.norm(g_acc_vector_spot[spot_index], axis=1)[spot.faces], axis=1))
            setattr(spot, 'log_g', np.log10(spot.potential_gradient_magnitudes))

    return system_container


def calculate_polar_potential_gradient_magnitude(polar_radius, mass):
    """
    Returns magnitude of polar gradient of gravitational potential.

    :param polar_radius: float;
    :param mass: float;
    :return: float;
    """
    points_z = polar_radius
    r3 = np.power(points_z, 3)
    domega_dz = c.G * mass * points_z / r3
    return domega_dz


def calculate_potential_gradient(points, angular_velocity, mass):
    """
    Returns array of gravity potential gradients for corresponding faces.

    :param points: numpy.array; (N * 3) array of surface points
    :param angular_velocity: float; angular velocity of rotation
    :param mass: float; stellar mass
    :return: numpy.array; gravity gradients for each face
    """
    r3 = np.power(np.linalg.norm(points, axis=1), 3)
    points_gradients = np.empty(points.shape)
    points_gradients[:, 0] = c.G * mass * points[:, 0] / r3 - np.power(angular_velocity, 2) * points[:, 0]
    points_gradients[:, 1] = c.G * mass * points[:, 1] / r3 - np.power(angular_velocity, 2) * points[:, 1]
    points_gradients[:, 2] = c.G * mass * points[:, 2] / r3

    return - points_gradients



from .. import units as u
from .. base.transform import SystemProperties, WHEN_FLOAT64, quantity_transform


class SingleSystemProperties(SystemProperties):
    @staticmethod
    def rotation_period(value):
        """
        Transform and validate rotational period of star in single star system, if unit is not specified, default period
        unit is assumed
        :param value: quantity or float; rotation period
        :return: float
        """
        return quantity_transform(value, u.DefaultSingleSystemUnits.system.rotation_period, WHEN_FLOAT64,
                                  u.DefaultSingleSystemInputUnits.system.rotation_period)

    @staticmethod
    def reference_time(value):
        """
        Transform and validity check for reference time.

        :param value: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]
        :return: float
        """
        return quantity_transform(value, u.DefaultSingleSystemUnits.system.reference_time, WHEN_FLOAT64,
                                  u.DefaultSingleSystemInputUnits.system.reference_time)



import numpy as np

from ... import settings
from .. utils import generate_phase_shift


def calculate_horizontal_displacements(mode, thetas, harmonics_derivatives, radius, scale):
    """
    Calculate angular horizontal components of displacement.

    :param scale:
    :param mode: PulsationMode;
    :param thetas: numpy.array; 1D
    :param harmonics_derivatives: numpy.array; column-wise - dY/dphi, dY/dtheta
    :param radius:
    :return:
    """
    if mode.l == 0:
        return np.zeros(thetas.shape[0]), np.zeros(thetas.shape[0])
    # TODO: remove this sin
    sin_theta = np.sin(thetas)
    theta_test = sin_theta != 0

    # lambda - distance along phi
    # nu - distance along theta coordinate
    phi_displacement = np.zeros(thetas.shape, dtype=np.complex128)
    phi_displacement[theta_test] = harmonics_derivatives[0, theta_test] / np.power(sin_theta[theta_test], 2)

    d_lamda = radius * sin_theta * np.abs(phi_displacement)
    theta_amp = np.abs(harmonics_derivatives[1])
    d_nu = radius * theta_amp

    dr = np.sqrt(np.mean(np.power(d_lamda, 2) + np.power(d_nu, 2)))
    corr_factor = mode.horizontal_amplitude / (dr * scale)

    phi_retval = np.zeros(thetas.shape, dtype=np.complex128)
    phi_retval[theta_test] = corr_factor * phi_displacement[theta_test]
    return phi_retval, corr_factor * harmonics_derivatives[1]


# ________________________complex angular coordinates_______________________
def calculate_displacement_coordinates(mode, points, harmonics, harmonics_derivatives, radius, scale=1.0):
    """
    Calculates surface displacement caused by given `mode`.

    :param mode: elisa.pulse.mode.Mode;
    :param points: numpy.array; in spherical coordinates
    :param harmonics: numpy.array; Y_l^m
    :param harmonics_derivatives: numpy.array; [dY/dphi, dY/dtheta]
    :param radius: float; equivalent radius of the component
    :param scale: numpy.float; scale of the perturbations
    :return: numpy.array; complex
    """
    if settings.PULSATION_MODEL == 'uniform':
        radial_displacement = calculate_radial_displacement(mode, harmonics) / scale
        phi_displacement, theta_displacement = \
            calculate_horizontal_displacements(mode, points[:, 2], harmonics_derivatives, radius, scale)

        return np.column_stack((radial_displacement, phi_displacement, theta_displacement))
    else:
        raise NotImplementedError(f'Pulsation model: {settings.PULSATION_MODEL} is not implemented.')


def calculate_mode_angular_displacement(displacement):
    """
    Calculates angular displacement from complex angular coordinates.

    :param displacement: numpy.array; complex
    :return: numpy.array;
    """
    return np.real(displacement)


def calculate_radial_displacement(mode, harmonics):
    """
    Calculates radial displacement of surface points.

    :param mode: PulsationMode;
    :param harmonics: numpy.array; Y_l^m
    :return: numpy.array; complex
    """
    return mode.radial_amplitude * harmonics


# ________________________velocity coordinates_______________________
def calculate_mode_derivatives(displacement, angular_frequency):
    """
    Calculates derivatives of angular displacement.

    :param displacement: numpy.array; complex
    :param angular_frequency: np.float;
    :return: numpy.array;
    """
    return angular_frequency * np.imag(displacement)


# _______________________acceleration coordinates_______________________
def calculate_mode_second_derivatives(displacement, angular_frequency):
    """
    Calculates derivatives of angular displacement.

    :param displacement: numpy.array; complex
    :param angular_frequency: np.float;
    :return: numpy.array;
    """
    return - angular_frequency**2 * np.real(displacement)


# _______________________temperature_perturbation_______________________
def calculate_temperature_pert_factor(mode, scale):
    """
    Returns perturbation factor (delta T = T_factor * T) for surface temperature based on a treatment in Townsend 2003.

    :param mode: PulsationMode;
    :param scale: float; system scale
    :return: numpy.array;
    """
    hrm_shift = np.real(
        generate_phase_shift(mode.temperature_perturbation_phase_shift) * mode.complex_displacement[:, 0]
    )
    return mode.temperature_amplitude_factor * hrm_shift * scale / mode.radial_amplitude



import functools
import shutil
import tempfile
import requests
import tarfile

from tqdm import tqdm

from typing import Iterable
from logging import getLogger

logger = getLogger("elisa.managers.download_manager")


class DownloadManager(object):
    def __init__(self, settings):
        self.settings = settings

    @staticmethod
    def temp_file():
        return tempfile.NamedTemporaryFile().name

    @staticmethod
    def extract(archive_path, destination_path, truncate_tarfile_path=""):
        def truncate_members(_archive) -> Iterable[tarfile.TarInfo]:
            truncate_length = len(truncate_tarfile_path)
            for member in _archive.getmembers():
                if member.path.startswith(truncate_tarfile_path):
                    member.path = member.path[truncate_length:]
                    yield member

        logger.info(f"extract files to {destination_path}")
        with tarfile.open(archive_path, "r") as archive:
            import os

            members = [x for x in truncate_members(archive)]

            def is_within_directory(directory, target):
                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)
                prefix = os.path.commonprefix([abs_directory, abs_target])
                return prefix == abs_directory

            def safe_extract(tar, path=".", members_=None, *, numeric_owner=False):
                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")
                tar.extractall(path, members_, numeric_owner=numeric_owner)
                
            safe_extract(archive, destination_path, members_=members)

    @classmethod
    def _download_and_process(cls, url, destination_path, truncate_tarfile_path=""):
        temp_path = cls.temp_file()
        cls.download_file(url, temp_path)
        cls.extract(temp_path, destination_path, truncate_tarfile_path=truncate_tarfile_path)

    def download_ck04(self):
        url = "https://github.com/mikecokina/elisa-assets/raw/refs/heads/main/atmosphere/ck04.tar.gz"
        logger.info("downloading castelli-kurucz 2004 atmosphere files")
        self._download_and_process(url, self.settings.CK04_ATM_TABLES, truncate_tarfile_path="ck04/")

    def download_k93(self):
        url = "https://github.com/mikecokina/elisa-assets/raw/refs/heads/main/atmosphere/k93.tar.gz"
        logger.info("downloading kurucz 1993 atmosphere files")
        self._download_and_process(url, self.settings.K93_ATM_TABLES, truncate_tarfile_path="k93/")

    def download_van_hamme(self):
        url = "https://github.com/mikecokina/elisa-assets/raw/refs/heads/main/limbdarkening/ld_vh19.tar.gz"
        logger.info("downloading van hamme 2019 limb darkening files")
        self._download_and_process(url, self.settings.LD_TABLES, truncate_tarfile_path="ld/")

    def download_all(self):
        logger.info("Download manager is running...")

        self.download_van_hamme()
        self.download_ck04()
        self.download_k93()

    @staticmethod
    def handler(file_size, url, path):
        headers = {'Range': f'bytes=0-{file_size}'}

        # Make GET request and allow redirects
        r = requests.get(url, headers=headers, stream=True, allow_redirects=True)

        # Ensure response is successful
        if r.status_code >= 400:
            raise Exception(f"Failed to download file: {url} (Status Code: {r.status_code})")

        # Handle decompression if needed
        r.raw.read = functools.partial(r.raw.read, decode_content=True)

        # Download with tqdm progress bar
        with tqdm.wrapattr(r.raw, "read", total=file_size, desc="Downloading") as r_raw:
            with open(path, "wb") as f:
                shutil.copyfileobj(r_raw, f)

    @classmethod
    def download_file(cls, url, path):
        # Allow redirects in HEAD request to get the final URL
        r = requests.head(url, headers={'Accept-Encoding': None}, allow_redirects=True)

        # Ensure response is successful
        if r.status_code >= 400:
            raise Exception(f"Failed to access URL: {url} (Status Code: {r.status_code})")

        # Get the final URL after redirections
        final_url = r.url

        # Get the file size
        file_size = int(r.headers.get('content-length', 0))  # Use .get() to avoid KeyError if missing

        if file_size == 0:
            raise Exception("Failed to determine file size. Content-Length missing.")

        cls.handler(file_size, final_url, path)



import numpy as np

# keept order of allowed constraints method or came with better way how to validaite and evaluate fitting constraints
ALLOWED_CONSTRAINT_METHODS = ['arcsin', 'arccos', 'arctan', 'log', 'sin', 'cos', 'tan', 'exp', 'degrees', 'radians']
ALLOWED_CONSTRAINT_CHARS = ['(', ')', '+', '-', '*', '/', '.', 'e'] + [str(i) for i in range(0, 10, 1)]
TRANSFORM_TO_METHODS = ['arcs', 'arcc', 'arct', 'log', 'sin', 'cos', 'tan', 'exp', 'deg', 'rad']

arcs = np.arcsin
arcc = np.arccos
arct = np.arctan

sin = np.sin
cos = np.cos
tan = np.tan

exp = np.exp
log = np.log
log10 = np.log10

deg = np.degrees
rad = np.radians



import numpy as np

from . mixins import MCMCMixin
from . shared import AbstractFit

from .. params import parameters
from .. params.parameters import ParameterMeta, BinaryInitialParameters


def filter_chain(mcmc_fit_cls, **boundaries):
    """
    Filtering mcmc chain down to given parameter intervals. This function is usable in case of bimodal distribution of
    the MCMC chain.

    :param mcmc_fit_cls: MCMC fitting cls instance e.g.: (LCFitMCMC, RVFitMCMC)
    :param boundaries: Dict; dictionary of boundaries in flat format (using @)
                             e.g. {`primary@te_ff`: (5000, 6000), other parameters ...}
    :return: numpy.array; filtered flat chain
    """
    for key, boundary in boundaries.items():
        if not isinstance(boundary, (tuple, list, np.ndarray)):
            raise TypeError(f'`{key}` boundary is not tuple or list.')
        if len(boundary) != 2:
            raise TypeError(f'`{key}` has incorrect length of {len(boundary)}.')
        if key not in mcmc_fit_cls.variable_labels:
            raise NameError(f'{key} is not valid model parameter.')

        column_idx = mcmc_fit_cls.variable_labels.index(key)
        column = mcmc_fit_cls.flat_chain[:, column_idx]

        condition_mask = np.logical_and(
            column > parameters.normalize_value(boundary[0], *mcmc_fit_cls.normalization[key]),
            column < parameters.normalize_value(boundary[1], *mcmc_fit_cls.normalization[key])
        )
        if np.sum(condition_mask) == 0:
            raise ValueError(f'Boundaries for {key} yielded an empty array.')
        setattr(mcmc_fit_cls, 'flat_chain', mcmc_fit_cls.flat_chain[condition_mask, :])

    fitted_params = {key: mcmc_fit_cls.flat_result[key] for key in mcmc_fit_cls.variable_labels}
    update_solution(mcmc_fit_cls, fitted_params, percentiles=None)

    return mcmc_fit_cls.flat_chain


def load_chain(mcmc_fit_cls, fit_id, discard=0, percentiles=None):
    """
    Function loads MCMC chain along with auxiliary data from json file created after each MCMC run.

    :param percentiles: List; percentile intervals used to generate confidence intervals, provided in form:
                              [percentile for lower bound of confidence interval, percentile of the centre,
                              percentile for the upper bound of confidence interval]
    :param mcmc_fit_cls: MCMC fitting cls instance e.g.: (LCFitMCMC, RVFitMCMC)
    :param discard: int; Discard the first discard steps in the chain as burn-in. (default: 0)
    :param fit_id: str; chain identificator or filename containing the chain
    :return: Tuple[numpy.ndarray, list, dict]; flattened mcmc chain, labels of variables in `flat_chain` columns,
                                               {var_name: (min_boundary, max_boundary), ...} dictionary of
                                               boundaries defined by user for each variable needed
                                               to reconstruct real values from normalized `flat_chain` array
    """
    data = MCMCMixin.load_flat_chain(fit_id=fit_id)

    mcmc_fit_cls.flat_chain = np.array(data['flat_chain'])[discard:, :]
    mcmc_fit_cls.variable_labels = data['fitable_parameters']
    mcmc_fit_cls.normalization = data['normalization']

    update_solution(mcmc_fit_cls, data['fitable'], percentiles)

    return mcmc_fit_cls.flat_chain, mcmc_fit_cls.variable_labels, mcmc_fit_cls.normalization


def update_solution(mcmc_fit_cls, fitted_params, percentiles):
    """
    Updating solutions based on the distribution of to MCMC chain.

    :param mcmc_fit_cls: MCMC fitting cls instance e.g.: (LCFitMCMC, RVFitMCMC)
    :param fitted_params: Dict; only variable part of flat_result
    :param percentiles: List; percentiles used for evaluation of confidence intervals
    :return: Tuple;
    """
    fitable = {key: ParameterMeta(**val) for key, val in fitted_params.items()}

    # reproducing results from chain
    flat_result_update = MCMCMixin.resolve_mcmc_result(mcmc_fit_cls.flat_chain, fitable,
                                                       mcmc_fit_cls.normalization, percentiles=percentiles)

    if mcmc_fit_cls.result is not None:
        mcmc_fit_cls.flat_result.update(flat_result_update)

        # evaluating constraints
        fit_params = parameters.serialize_result(mcmc_fit_cls.flat_result)
        constrained = BinaryInitialParameters(**fit_params).get_constrained()
        mcmc_fit_cls.flat_result = AbstractFit.eval_constrained_results(mcmc_fit_cls.flat_result, constrained)

        mcmc_fit_cls.result = parameters.serialize_result(mcmc_fit_cls.flat_result)

    else:
        msg = 'Load fit parameters before loading the chain. For eg. UPDATE THIS.'
        raise ValueError(msg)


def write_ln(write_fn, designation, value, bot, top, unit, status, line_sep, precision=8):
    val = round(value, precision) if type(value) is not str else value
    return write_fn(f"{designation:<35} "
                    f"{val:>20}"
                    f"{bot:>20}"
                    f"{top:>20}"
                    f"{unit:>20}    "
                    f"{status:<50}{line_sep}")


def write_param_ln(fit_params, param_id, designation, write_fn, line_sep, precision=8):
    """
    Auxiliary function to the fit_summary functions, produces one
    line in output for given parameter that is present in `fit_params`.

    :param precision: int;
    :param fit_params: Dict;
    :param param_id: str; name os the parameter in `fit_params`
    :param designation: str; displayed name of the parameter
    :param write_fn: function used to write into console or to the file
    :param line_sep: str; symbols to finish the line
    :return:
    """

    if 'confidence_interval' in fit_params[param_id]:
        bot = fit_params[param_id]['value'] - fit_params[param_id]['confidence_interval']['min']
        top = fit_params[param_id]['confidence_interval']['max'] - fit_params[param_id]['value']

        aux = np.abs([bot, top])
        aux[aux == 0] = 1e6
        sig_figures = -int(np.log10(np.min(aux))//1) + 1

        bot = round(bot, sig_figures)
        top = round(top, sig_figures)
    else:
        bot, top = '-', '-',
        sig_figures = precision

    status = 'Not recognized'
    if 'fixed' in fit_params[param_id]:
        status = 'Fixed' if fit_params[param_id]['fixed'] else 'Variable'

    elif 'constraint' in fit_params[param_id].keys():
        status = fit_params[param_id]['constraint']
    elif param_id in ['r_squared']:
        status = 'Derived'

    unit = str(fit_params[param_id]['unit']) if 'unit' in fit_params[param_id].keys() else '-'
    args = write_fn, designation, round(fit_params[param_id]['value'], sig_figures), bot, top, unit, status, line_sep
    return write_ln(*args)


def write_propagated_ln(values, fit_params, param_id, designation, write_fn, line_sep, unit):
    """
    Auxiliary function to the fit_summary functions, produces one
    line in output for given parameter that is present in `fit_params`.

    :param unit:
    :param values:
    :param fit_params: Dict;
    :param param_id: str; name os the parameter in `fit_params`
    :param designation: str; displayed name of the parameter
    :param write_fn: function used to write into console or to the file
    :param line_sep: str; symbols to finish the line
    :return:
    """
    # if parameter does not exist in given fitting mode, the line in summary is omitted
    if np.isnan(values).any():
        return

    aux = np.abs([values[1], values[2]])
    aux[aux <= 1e-15] = 1e-15
    sig_figures = -int(np.log10(np.min(aux))//1) + 1

    values = np.round(values, sig_figures)

    if param_id not in fit_params.keys():
        status = 'Derived'
    elif 'fixed' in fit_params[param_id]:
        status = 'Fixed' if fit_params[param_id]['fixed'] else 'Variable'

    elif 'constraint' in fit_params[param_id].keys():
        status = fit_params[param_id]['constraint']
    elif param_id in ['r_squared']:
        status = 'Derived'
    else:
        status = 'Unknown'

    return write_ln(write_fn, designation, values[0], values[1], values[2], unit, status, line_sep)



import numpy as np
import matplotlib.path as mpltpath

from scipy.spatial.qhull import ConvexHull
from copy import copy

from .. import utils as bsutils
from ... import settings
from ... base.types import FLOAT, INT
from ... logger import getLogger
from ... import (
    umpy as up,
    utils,
    const
)
from ... base.surface import coverage as bcoverage


logger = getLogger('binary_system.surface.coverage')


def partial_visible_faces_surface_coverage(points, faces, normals, hull):
    """
    Compute surface coverage of partialy visible faces.

    :param points: numpy.array;
    :param faces: numpy.array;
    :param normals: numpy.array;
    :param hull: numpy.array; sorted clockwise to create
                              matplotlib.path.Path; path of points boundary of infront component projection
    :return: numpy.array;
    """
    pypex_hull = bsutils.hull_to_pypex_poly(hull)
    pypex_faces = bsutils.faces_to_pypex_poly(points[faces])
    # it is possible to None happens in intersection, tkae care about it latter
    pypex_intersection = bsutils.pypex_poly_hull_intersection(pypex_faces, pypex_hull)

    # think about surface normalisation like and avoid surface areas like 1e-6 which lead to loss in precission
    pypex_polys_surface_area = np.array(bsutils.pypex_poly_surface_area(pypex_intersection), dtype=FLOAT)

    inplane_points_3d = np.column_stack((points, np.zeros(points.shape[0])))
    inplane_surface_area = utils.triangle_areas(triangles=faces, points=inplane_points_3d)
    correction_cosine = utils.calculate_cos_theta_los_x(normals)
    retval = (inplane_surface_area - pypex_polys_surface_area) / correction_cosine
    return retval


def calculate_centre_of_star_projection(system, component):
    """
    Returns yz projection of centre of mass of given `component`.

    :param system: elisa.binary_system.container.OrbitalPositionContainer
    :param component: str; `primary` or `secondary`
    :return: numpy.array;
    """
    if component == 'primary':
        return np.array([0.0, 0.0])
    else:
        centre_vector = np.array([system.position.distance, 0.0, 0.0])
        args = (system.position.azimuth - const.HALF_PI, centre_vector, "z", False, False)
        centre_vector = utils.around_axis_rotation(*args)

        args = (const.HALF_PI - system.inclination, centre_vector, "y", False, False)
        centre_vector = utils.around_axis_rotation(*args)

        return centre_vector[1:]


def expand_star_outline(path, system, cover_component):
    """
    Function takes outline of the cover star and expands it slightly to compensate for the loss of its area due to
    surface discretization.

    :param path: Path; outline of the eclipsing star
    :param system: elisa.binary_system.container.OrbitalPositionContainer
    :param cover_component: str; `primary` or `secondary`
    :return: Path; expanded outline of the star
    """
    cpm = calculate_centre_of_star_projection(system, cover_component)
    alpha = const.FULL_ARC / np.shape(path.vertices)[0]
    corr_factor = np.sqrt(2 - (np.sin(alpha) / alpha))
    path.vertices = corr_factor * (path.vertices - cpm[None, :]) + cpm[None, :]
    return path


def test_size_similarity(cover_object, undercover_object):
    """
    Checking whether size of the cover component is comparable to the triangle size of the undercovar component which
    requires a separate approach.

    :param cover_object: StarContainer;
    :param undercover_object: StarContainer;
    :return: bool;
    """
    cover_size = 2.0 * cover_object.equivalent_radius
    undercover_triangle_size = undercover_object.equivalent_radius * np.sin(undercover_object.discretization_factor)
    return cover_size <= undercover_triangle_size


def visibility_out_of_eclipse(undercover_object):
    """
    Decides visibility of the near side faces of the undercover component (eclipsed component) outside of eclipse.

    :param undercover_object: StarContainer; eclipsed star
    :return: tuple; full_visible, invisible, partial_visible triangles observer-facing part of eclipsed component
    """
    eclipse_faces_visibility = np.full(undercover_object.normals.shape, False, dtype=bool)
    eclipse_faces_visibility[undercover_object.indices] = True

    # get indices of full visible, invisible and partial visible faces
    full_visible = np.all(eclipse_faces_visibility, axis=1)
    placeholder = np.full(full_visible.shape, False, dtype=bool)
    return full_visible, placeholder, placeholder


def visibility_similar_objects(undercover_visible_projection, undercover_object, undercover_visible_point_indices,
                               cover_outline):
    """
    Decides visibility of the near side faces of the undercover component (eclipsed component) during eclipse in case
    of similarly sized components where cover (eclipsing) component is much larger than triangle on eclipsed component.

    :param undercover_visible_projection: numpy.array; observer-facing points of eclipsed component
    :param undercover_object: StarContainer; eclipsed component
    :param undercover_visible_point_indices: numpy.array; indices of observer-facing points of eclipsed component
    :param cover_outline: matplotlib.path.Path; hull of the eclipsing component
    :return: tuple; full_visible, invisible, partial_visible triangles observer-facing part of eclipsed component
    """
    # obtain points out of eclipse (out of boundary defined by hull of 'infront' object)
    out_of_bound = up.invert(cover_outline.contains_points(undercover_visible_projection))

    undercover_visible_point_indices = undercover_visible_point_indices[out_of_bound]
    undercover_faces = np.full(undercover_object.normals.shape, -1, dtype=INT)
    undercover_faces[undercover_object.indices] = undercover_object.faces[undercover_object.indices]

    eclipse_faces_visibility = np.isin(undercover_faces, undercover_visible_point_indices)

    # get indices of full visible, invisible and partial visible faces
    full_visible = np.all(eclipse_faces_visibility, axis=1)
    invisible = np.all(up.invert(eclipse_faces_visibility), axis=1)
    partial_visible = up.invert(full_visible | invisible)
    return full_visible, invisible, partial_visible


def visibility_disimilar_objects(undercover_visible_projection, undercover_object, undercover_visible_point_indices,
                                 cover_outline):
    """
    Decides visibility of the near side faces of the undercover component (eclipsed component) during eclipse in case
    of significantly smaller (eclipsing) component that is comparable or smaller to triangle on eclipsed component.

    :param undercover_visible_projection: numpy.array; observer-facing points of eclipsed component
    :param undercover_object: StarContainer; eclipsed component
    :param undercover_visible_point_indices: numpy.array; indices of observer-facing points of eclipsed component
    :param cover_outline: matplotlib.path.Path; hull of the eclipsing component
    :return: tuple; full_visible, invisible, partial_visible triangles observer-facing part of eclipsed component
    """
    outline_max_coord = cover_outline.vertices.max(axis=0)
    outline_min_coord = cover_outline.vertices.min(axis=0)
    cover_centre = 0.5 * (outline_max_coord + outline_min_coord)  # centre of eclipsed component
    selection_radius = undercover_object.equivalent_radius * np.sin(undercover_object.discretization_factor)

    # square searchbox around cover component COM with half size equivalent to triangle size
    # noinspection PyUnresolvedReferences
    max_condition = np.array((undercover_visible_projection < (cover_centre + selection_radius)[None, :])).all(axis=1)
    min_condition = np.array((undercover_visible_projection > (cover_centre - selection_radius)[None, :])).all(axis=1)

    out_of_bound = ~np.logical_and(max_condition, min_condition)

    undercover_visible_point_indices = undercover_visible_point_indices[out_of_bound]
    undercover_faces = np.full(undercover_object.normals.shape, -1, dtype=INT)
    undercover_faces[undercover_object.indices] = undercover_object.faces[undercover_object.indices]

    eclipse_faces_visibility = np.isin(undercover_faces, undercover_visible_point_indices)

    full_visible = np.all(eclipse_faces_visibility, axis=1)
    invisible = np.full(full_visible.shape, False, dtype=bool)
    partial_visible = copy(full_visible)
    partial_visible[undercover_object.indices] = ~partial_visible[undercover_object.indices]
    return full_visible, invisible, partial_visible


def compute_surface_coverage(system, semi_major_axis, in_eclipse=True, return_values=True, write_to_containers=False):
    # todo: add unittests
    """
    Compute surface coverage of faces for given orbital position
    defined by container/SingleOrbitalPositionContainer.

    :param semi_major_axis: float;
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param in_eclipse: bool;
    :param return_values: bool; return coverages
    :param write_to_containers: bool; calculated values will be assigned to `system` container
    :return: Dict;
    """
    logger.debug(f"computing surface coverage for {system.position}")
    cover_component = 'secondary' if 0.0 < system.position.azimuth < const.PI else 'primary'
    cover_object = getattr(system, cover_component)
    undercover_object = getattr(system, settings.BINARY_COUNTERPARTS[cover_component])

    # all surface values in sma unit which are smaller then following threshold are discarded (set to 0.0)
    surface_noise_threshold = (2.0 * np.pi * np.power(undercover_object.polar_radius, 2) /
                               len(undercover_object.faces)) / 1e6

    cover_object_obs_visible_projection = utils.get_visible_projection(cover_object)
    undercover_object_obs_visible_projection = utils.get_visible_projection(undercover_object)

    if in_eclipse:
        # indices of points on near side
        undercover_visible_point_indices = np.unique(undercover_object.faces[undercover_object.indices])

        # outline of the eclipsing component
        bb_path = get_eclipse_boundary_path(cover_object_obs_visible_projection)

        similar_size_test = test_size_similarity(cover_object, undercover_object)
        args = (undercover_object_obs_visible_projection, undercover_object, undercover_visible_point_indices, bb_path)
        full_visible, invisible, partial_visible = \
            visibility_disimilar_objects(*args) if similar_size_test else visibility_similar_objects(*args)
    else:
        full_visible, invisible, partial_visible = visibility_out_of_eclipse(undercover_object)

    # process partial and full visible faces (get surface area of 3d polygon) of undercover object
    partial_visible_faces = undercover_object.faces[partial_visible]
    partial_visible_normals = undercover_object.normals[partial_visible]
    undercover_object_pts_projection = utils.plane_projection(undercover_object.points, "yz", keep_3d=False)
    if in_eclipse:
        # noinspection PyUnboundLocalVariable
        partial_coverage = partial_visible_faces_surface_coverage(
            points=undercover_object_pts_projection,
            faces=partial_visible_faces,
            normals=partial_visible_normals,
            hull=bb_path.vertices
        )
        partial_coverage[partial_coverage < surface_noise_threshold] = 0.0
    else:
        partial_coverage = None

    # discard values of surface which are under threshold
    visible_coverage = undercover_object.areas[full_visible]

    undercover_obj_coverage = bcoverage.surface_area_coverage(
        size=np.shape(undercover_object.normals)[0],
        visible=full_visible, visible_coverage=visible_coverage,
        partial=partial_visible, partial_coverage=partial_coverage
    )

    cover_obj_coverage = np.zeros(cover_object.areas.shape)
    cover_obj_coverage[cover_object.indices] = cover_object.areas[cover_object.indices]

    # areas are now in SMA^2, converting to SI
    cover_obj_coverage *= up.power(semi_major_axis, 2)
    undercover_obj_coverage *= up.power(semi_major_axis, 2)

    if write_to_containers:
        setattr(cover_object, 'coverage', cover_obj_coverage)
        setattr(undercover_object, 'coverage', undercover_obj_coverage)

    return {
        cover_component: cover_obj_coverage,
        settings.BINARY_COUNTERPARTS[cover_component]: undercover_obj_coverage
    } if return_values else None


def get_eclipse_boundary_path(hull):
    """
    Return `matplotlib.path.Path` object which represents boundary of component projection
    to plane `yz`.

    :param hull: numpy.array;
    :return: matplotlib.path.Path;
    """
    cover_bound = ConvexHull(hull)
    hull_points = hull[cover_bound.vertices]
    bb_path = mpltpath.Path(hull_points)
    return bb_path


def calculate_coverage_with_cosines(system, semi_major_axis, in_eclipse=True):
    """
    Function prepares surface-related parameters such as coverage(area of visibility
    of the triangles) and directional cosines towards line-of-sight vector.

    :param semi_major_axis: float;
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param in_eclipse: bool; indicate if eclipse occur for given position container.
                             If you are not sure leave it to True
    :return: Tuple;

    shape::

        (numpy.array, Dict[str, numpy.array])

    coverage -- numpy.array; visible area of triangles
    p_cosines, s_cosines -- Dict[str, numpy.array]; directional cosines for each face with
    respect to line-of-sight vector
    """
    coverage = compute_surface_coverage(system, semi_major_axis=semi_major_axis, in_eclipse=in_eclipse)
    p_cosines = utils.calculate_cos_theta_los_x(system.primary.normals)
    s_cosines = utils.calculate_cos_theta_los_x(system.secondary.normals)
    cosines = {'primary': p_cosines, 'secondary': s_cosines}
    return coverage, cosines
